import os
import numpy as np
import xml.etree.ElementTree as ET


def _get_voc_instances_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)
    instances = tree.findall('object')
    voc_instances = []
    for instance in instances:
        class_name = instance.find("name").text
        class_attributes = []
        for attr in ['pose', 'occluded', 'difficult', 'truncated']:
            attr_value = instance.find(attr)
            if attr_value is not None:
                class_attributes.append(
                    {
                        'name': attr_value.text,
                        'groupName': attr
                    }
                )

        bbox = instance.find("bndbox")
        bbox = [
            float(bbox.find(x).text) for x in ["xmin", "ymin", "xmax", "ymax"]
        ]
        voc_instances.append(({class_name: class_attributes}, bbox))
    return voc_instances


def _iou(bbox1, bbox2):
    xmin1, ymin1, xmax1, ymax1 = bbox1
    xmin2, ymin2, xmax2, ymax2 = bbox2

    x = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    y = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    return x * y / float(
        (xmax1 - xmin1) * (ymax1 - ymin1) + (xmax2 - xmin2) *
        (ymax2 - ymin2) - x * y
    )


def _get_image_shape_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)

    size = tree.find('size')
    width = int(size.find('width').text)
    height = int(size.find('height').text)

    return height, width



from mixpanel import Mixpanel
from .config import TOKEN

mp = Mixpanel(TOKEN)


def get_default(team_name, user_id, project_name=None):
    return {
        "SDK": True,
        "Paid": True,
        "Team": team_name,
        "Team Owner": user_id,
        "Project Name": project_name,
        "Project Role": "Admin",
    }



from pathlib import Path

import superannotate as sa

from ..common import upload_project


def test_coco_vector_instance(tmpdir):
    project_name = "coco2sa_vector_instance"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "instance_segmentation"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "COCO", "instances_test", "Vector",
        "instance_segmentation"
    )

    description = 'coco vector instance segmentation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_coco_vector_object(tmpdir):
    project_name = "coco2sa_vector_object"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "instance_segmentation"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "COCO", "instances_test", "Vector",
        "object_detection"
    )

    description = 'coco vector object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_coco_vector_keypoint(tmpdir):
    project_name = "coco2sa_keypoint"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "keypoint_detection/"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "COCO", "person_keypoints_test", "Vector",
        "keypoint_detection"
    )

    description = 'coco vector keypoint detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_coco_panoptic(tmpdir):
    project_name = "coco2sa_panoptic"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "panoptic_segmentation"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "COCO", "panoptic_test", "Pixel",
        "panoptic_segmentation"
    )

    description = 'coco pixel panoptic segmentation'
    ptype = 'Pixel'
    upload_project(out_dir, project_name, description, ptype)


def test_coco_pixel_instance(tmpdir):
    project_name = "coco2sa_pixel_instance"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "instance_segmentation"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "COCO", "instances_test", "Pixel",
        "instance_segmentation"
    )

    description = 'coco pixel instance segmentation'
    ptype = 'Pixel'
    upload_project(out_dir, project_name, description, ptype)


def test_sa_to_coco_to_sa(tmpdir):
    input_dir = Path("tests") / "sample_project_pixel"
    output1 = Path(tmpdir) / 'to_coco'
    output2 = Path(tmpdir) / 'to_sa'

    sa.export_annotation(
        input_dir, output1, "COCO", "object_test", "Pixel",
        "instance_segmentation"
    )

    sa.import_annotation(
        output1, output2, "COCO", "object_test", "Pixel",
        "instance_segmentation", 'image_set'
    )

    project_name = 'coco_pipeline_new'
    description = 'test_instane'
    ptype = 'Pixel'
    upload_project(output2, project_name, description, ptype)



'''
'''
import logging

from .coco_converters.coco_to_sa_pixel import (
    coco_instance_segmentation_to_sa_pixel,
    coco_panoptic_segmentation_to_sa_pixel
)
from .coco_converters.coco_to_sa_vector import (
    coco_instance_segmentation_to_sa_vector,
    coco_keypoint_detection_to_sa_vector, coco_object_detection_to_sa_vector
)
from .coco_converters.sa_pixel_to_coco import (
    sa_pixel_to_coco_instance_segmentation,
    sa_pixel_to_coco_panoptic_segmentation
)
from .coco_converters.sa_vector_to_coco import (
    sa_vector_to_coco_instance_segmentation,
    sa_vector_to_coco_keypoint_detection, sa_vector_to_coco_object_detection
)

from .voc_converters.voc_to_sa_pixel import voc_instance_segmentation_to_sa_pixel
from .voc_converters.voc_to_sa_vector import (
    voc_object_detection_to_sa_vector, voc_instance_segmentation_to_sa_vector
)

from .dataloop_converters.dataloop_to_sa_vector import dataloop_to_sa

from .labelbox_converters.labelbox_to_sa_vector import labelbox_to_sa
from .labelbox_converters.labelbox_to_sa_pixel import labelbox_instance_segmentation_to_sa_pixel

from .sagemaker_converters.sagemaker_to_sa_vector import sagemaker_object_detection_to_sa_vector
from .sagemaker_converters.sagemaker_to_sa_pixel import sagemaker_instance_segmentation_to_sa_pixel

from .supervisely_converters.supervisely_to_sa_vector import (
    supervisely_to_sa, supervisely_keypoint_detection_to_sa_vector
)

from .supervisely_converters.supervisely_to_sa_pixel import (
    supervisely_instance_segmentation_to_sa_pixel
)

from .vgg_converters.vgg_to_sa_vector import vgg_to_sa

from .vott_converters.vott_to_sa_vector import vott_to_sa

from .googlecloud_converters.googlecloud_to_sa_vector import googlecloud_to_sa_vector

from .yolo_converters.yolo_to_sa_vector import yolo_object_detection_to_sa_vector

logger = logging.getLogger("superannotate-python-sdk")

CONVERSION_ALGORITHMS = {
    'from':
        {
            'COCO':
                {
                    'Vector':
                        {
                            'keypoint_detection':
                                coco_keypoint_detection_to_sa_vector,
                            'instance_segmentation':
                                coco_instance_segmentation_to_sa_vector,
                            'object_detection':
                                coco_object_detection_to_sa_vector
                        },
                    'Pixel':
                        {
                            'panoptic_segmentation':
                                coco_panoptic_segmentation_to_sa_pixel,
                            'instance_segmentation':
                                coco_instance_segmentation_to_sa_pixel
                        }
                },
            'VOC':
                {
                    'Vector':
                        {
                            'object_detection':
                                voc_object_detection_to_sa_vector,
                            'instance_segmentation':
                                voc_instance_segmentation_to_sa_vector
                        },
                    'Pixel':
                        {
                            'instance_segmentation':
                                voc_instance_segmentation_to_sa_pixel
                        }
                },
            'LabelBox':
                {
                    'Vector':
                        {
                            'object_detection': labelbox_to_sa,
                            'instance_segmentation': labelbox_to_sa,
                            'vector_annotation': labelbox_to_sa
                        },
                    'Pixel':
                        {
                            'instance_segmentation':
                                labelbox_instance_segmentation_to_sa_pixel
                        }
                },
            'DataLoop':
                {
                    'Vector':
                        {
                            'object_detection': dataloop_to_sa,
                            'instance_segmentation': dataloop_to_sa,
                            'vector_annotation': dataloop_to_sa
                        },
                    'Pixel': {}
                },
            'Supervisely':
                {
                    'Vector':
                        {
                            'vector_annotation':
                                supervisely_to_sa,
                            'instance_segmentation':
                                supervisely_to_sa,
                            'object_detection':
                                supervisely_to_sa,
                            'keypoint_detection':
                                supervisely_keypoint_detection_to_sa_vector
                        },
                    'Pixel':
                        {
                            'instance_segmentation':
                                supervisely_instance_segmentation_to_sa_pixel
                        }
                },
            'VoTT':
                {
                    'Vector':
                        {
                            'instance_segmentation': vott_to_sa,
                            'object_detection': vott_to_sa,
                            'vector_annotation': vott_to_sa
                        },
                    'Pixel': {}
                },
            'SageMaker':
                {
                    'Vector':
                        {
                            'object_detection':
                                sagemaker_object_detection_to_sa_vector
                        },
                    'Pixel':
                        {
                            'instance_segmentation':
                                sagemaker_instance_segmentation_to_sa_pixel
                        }
                },
            'VGG':
                {
                    'Vector':
                        {
                            'object_detection': vgg_to_sa,
                            'instance_segmentation': vgg_to_sa,
                            'vector_annotation': vgg_to_sa
                        },
                    'Pixel': {}
                },
            'GoogleCloud':
                {
                    'Vector': {
                        'object_detection': googlecloud_to_sa_vector
                    },
                    'Pixel': {}
                },
            'YOLO':
                {
                    'Vector':
                        {
                            'object_detection':
                                yolo_object_detection_to_sa_vector
                        },
                    'Pixel': {}
                }
        },
    'to':
        {
            'COCO':
                {
                    'Vector':
                        {
                            'instance_segmentation':
                                sa_vector_to_coco_instance_segmentation,
                            'object_detection':
                                sa_vector_to_coco_object_detection,
                            'keypoint_detection':
                                sa_vector_to_coco_keypoint_detection
                        },
                    'Pixel':
                        {
                            'panoptic_segmentation':
                                sa_pixel_to_coco_panoptic_segmentation,
                            'instance_segmentation':
                                sa_pixel_to_coco_instance_segmentation
                        }
                },
        }
}


class baseStrategy():
    def __init__(self, args):
        if args.dataset_format not in ('COCO', 'VOC'):
            logger.warning(
                "Beta feature. %s to SuperAnnotate annotation format converter is in BETA state."
                % args.dataset_format
            )
        self.project_type = args.project_type
        self.dataset_name = args.dataset_name
        self.export_root = args.export_root
        self.output_dir = args.output_dir
        self.task = args.task
        self.direction = args.direction
        self.conversion_algorithm = CONVERSION_ALGORITHMS[self.direction][
            args.dataset_format][self.project_type][self.task]

        self.name = "%s %s converter" % (args.dataset_format, self.task)

        self.failed_conversion_cnt = 0

    def __str__(self):
        return '%s object' % (self.name)

    def set_output_dir(self, output_dir_):
        self.output_dir = output_dir_

    def set_export_root(self, export_dir):
        self.export_root = export_dir

    def set_dataset_name(self, dname):
        self.dataset_name = dname

    def increase_converted_count(self):
        self.failed_conversion_cnt = self.failed_conversion_cnt + 1

    def set_num_converted(self, num_converted_):
        self.num_converted = num_converted_



def get_project_name(project):
    project_name = ""
    if isinstance(project, dict):
        project_name = project['name']
    if isinstance(project, str):
        if '/' in project:
            project_name = project.split('/')[0]
        else:
            project_name = project
    return project_name


def get_team_metadata(*args, **kwargs):
    return {"event_name": "get_team_metadata", "properties": {}}


def invite_contributor_to_team(*args, **kwargs):
    admin = kwargs.get("admin", None)
    if not admin:
        admin = args[1:2]
    if admin:
        admin = "CUSTOM"
    else:
        admin = "DEFAULT"
    return {
        "event_name": "invite_contributor_to_team",
        "properties": {
            "Admin": admin
        }
    }


def delete_contributor_to_team_invitation(*args, **kwargs):
    return {
        "event_name": "delete_contributor_to_team_invitation",
        "properties": {}
    }


def search_team_contributors(*args, **kwargs):
    return {
        "event_name": "search_team_contributors",
        "properties":
            {
                "Email": bool(args[0:1] or kwargs.get("email", None)),
                "Name": bool(args[1:2] or kwargs.get("first_name", None)),
                "Surname": bool(args[2:3] or kwargs.get("last_name", None))
            }
    }


def search_projects(*args, **kwargs):
    project = kwargs.get("name", None)
    if not project:
        project_name = None
        project = args[0:1]
        if project:
            project_name = get_project_name(project[0])
    else:
        project_name = get_project_name(project)
    return {
        "event_name": "search_projects",
        "properties":
            {
                "Metadata":
                    bool(args[2:3] or kwargs.get("return_metadata", None)),
                "project_name":
                    project_name
            }
    }


def create_project(*args, **kwargs):
    project = kwargs.get("project_name", None)
    if not project:
        project = args[0]
    project_type = kwargs.get("project_type", None)
    if not project_type:
        project_type = args[2]
    return {
        "event_name": "create_project",
        "properties":
            {
                "Project Type": project_type,
                "project_name": get_project_name(project)
            }
    }


def create_project_from_metadata(*args, **kwargs):
    project = kwargs.get("project_metadata", None)
    if not project:
        project = args[0]
    return {
        "event_name": "create_project_from_metadata",
        "properties": {
            "project_name": get_project_name(project)
        }
    }


def clone_project(*args, **kwargs):
    project = kwargs.get("project_name", None)
    if not project:
        project = args[0]
    return {
        "event_name": "clone_project",
        "properties":
            {
                "Copy Classes":
                    bool(
                        args[3:4] or
                        kwargs.get("copy_annotation_classes", None)
                    ),
                "Copy Settings":
                    bool(args[4:5] or kwargs.get("copy_settings", None)),
                "Copy Workflow":
                    bool(args[5:6] or kwargs.get("copy_workflow", None)),
                "Copy Contributors":
                    bool(args[6:7] or kwargs.get("copy_contributors", None)),
                "project_name":
                    get_project_name(project)
            }
    }


def search_images(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "search_images",
        "properties":
            {
                "Annotation Status":
                    bool(args[2:3] or kwargs.get("annotation_status", None)),
                "Metadata":
                    bool(args[3:4] or kwargs.get("return_metadata", None)),
                "project_name":
                    get_project_name(project)
            }
    }


def upload_images_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    img_paths = kwargs.get("img_paths", [])
    if not img_paths:
        img_paths += args[1]
    return {
        "event_name": "upload_images_to_project",
        "properties":
            {
                "Image Count":
                    len(img_paths),
                "Annotation Status":
                    bool(args[2:3] or kwargs.get("annotation_status", None)),
                "From S3":
                    bool(args[3:4] or kwargs.get("from_s3", None)),
                "project_name":
                    get_project_name(project)
            }
    }


def upload_image_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "upload_image_to_project",
        "properties":
            {
                "Image Name":
                    bool(args[2:3] or kwargs.get("image_name", None)),
                "Annotation Status":
                    bool(args[3:4] or kwargs.get("annotation_status", None)),
                "project_name":
                    get_project_name(project)
            }
    }


def upload_images_from_public_urls_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    img_urls = kwargs.get("img_urls", [])
    if not img_urls:
        img_urls += args[1]
    return {
        "event_name": "upload_images_from_public_urls_to_project",
        "properties":
            {
                "Image Count":
                    len(img_urls),
                "Image Name":
                    bool(args[2:3] or kwargs.get("img_names", None)),
                "Annotation Status":
                    bool(args[3:4] or kwargs.get("annotation_status", None)),
                "project_name":
                    get_project_name(project)
            }
    }


def upload_images_from_google_cloud_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "upload_images_from_google_cloud_to_project",
        "properties": {
            "project_name": get_project_name(project)
        }
    }


def upload_images_from_azure_blob_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "upload_images_from_azure_blob_to_project",
        "properties": {
            "project_name": get_project_name(project)
        }
    }


def upload_video_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]

    return {
        "event_name": "upload_video_to_project",
        "properties":
            {
                "project_name": get_project_name(project),
                "FPS": bool(args[2:3] or kwargs.get("target_fps", None)),
                "Start": bool(args[3:4] or kwargs.get("start_time", None)),
                "End": bool(args[4:5] or kwargs.get("end_time", None))
            }
    }


def attach_image_urls_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "attach_image_urls_to_project",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Image Names":
                    bool(args[1:2] or kwargs.get("attachments", None)),
                "Annotation Status":
                    bool(args[2:3] or kwargs.get("annotation_status", None))
            }
    }


def set_images_annotation_statuses(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    annotation_status = kwargs.get("annotation_status", None)
    if not annotation_status:
        annotation_status = args[2]
    image_names = kwargs.get("image_names", [])
    if not image_names:
        image_names = args[1]
    return {
        "event_name": "set_images_annotation_statuses",
        "properties":
            {
                "project_name": get_project_name(project),
                "Image Count": len(image_names),
                "Annotation Status": annotation_status
            }
    }


def get_image_annotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_image_annotations",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_image_preannotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_image_preannotations",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def download_image_annotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_image_annotations",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def download_image_preannotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_image_preannotations",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_image_metadata(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_image_metadata",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_image_bytes(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_image_bytes",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def delete_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "delete_image",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def add_annotation_comment_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_comment_to_image",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def delete_annotation_class(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "delete_annotation_class",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_annotation_class_metadata(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_annotation_class_metadata",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def download_annotation_classes_json(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_annotation_classes_json",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def search_annotation_classes(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "search_annotation_classes",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def unshare_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "unshare_project",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_image_count(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_project_image_count",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_settings(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_project_settings",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def set_project_settings(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "set_project_settings",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_default_image_quality_in_editor(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_project_default_image_quality_in_editor",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_metadata(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_project_metadata",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def delete_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "delete_project",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def rename_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "rename_project",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_workflow(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]

    return {
        "event_name": "get_project_workflow",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def set_project_workflow(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "set_project_workflow",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def create_folder(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "create_folder",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_folder_metadata(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_folder_metadata",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def get_project_and_folder_metadata(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_project_and_folder_metadata",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def rename_folder(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "rename_folder",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def stop_model_training(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "stop_model_training",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def download_model(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_model",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def plot_model_metrics(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "plot_model_metrics",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def delete_model(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "delete_model",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def convert_project_type(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "convert_project_type",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def convert_json_version(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "convert_json_version",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def df_to_annotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "df_to_annotations",
        "properties": {
            "project_name": get_project_name(project),
        }
    }


def upload_image_annotations(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "upload_image_annotations",
        "properties":
            {
                "project_name": get_project_name(project),
                "Pixel": bool(args[3:4] or ("mask" in kwargs)),
            }
    }


def download_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Download Annotations":
                    bool(args[3:4] or ("include_annotations" in kwargs)),
                "Download Fuse":
                    bool(args[4:5] or ("include_fuse" in kwargs)),
                "Download Overlay":
                    bool(args[5:6] or ("include_overlay" in kwargs)),
            }
    }


def copy_image(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "copy_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Copy Annotations":
                    bool(args[3:4] or ("include_annotations" in kwargs)),
                "Copy Annotation Status":
                    bool(args[4:5] or ("copy_annotation_status" in kwargs)),
                "Copy Pin":
                    bool(args[5:6] or ("copy_pin" in kwargs)),
            }
    }


def run_prediction(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    from superannotate.db.projects import get_project_metadata as sa_get_project_metadata
    project_name = get_project_name(project)
    project_metadata = sa_get_project_metadata(project_name)
    image_list = kwargs.get("images_list", None)
    if not image_list:
        image_list = args[1]
    return {
        "event_name": "run_prediction",
        "properties":
            {
                "Project Type": project_metadata['type'],
                "Image Count": len(image_list)
            }
    }


def run_segmentation(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    from superannotate.db.projects import get_project_metadata as sa_get_project_metadata
    project_name = get_project_name(project)
    project_metadata = sa_get_project_metadata(project_name)
    image_list = kwargs.get("images_list", None)
    if not image_list:
        image_list = args[1]
    model = kwargs.get("model", None)
    if not model:
        model = args[2]
    return {
        "event_name": "run_segmentation",
        "properties":
            {
                "Project Type": project_metadata['type'],
                "Image Count": len(image_list),
                "Model": model
            }
    }


def upload_videos_from_folder_to_project(*args, **kwargs):
    folder_path = kwargs.get("folder_path", None)
    if not folder_path:
        folder_path = args[1]
    from pathlib import Path
    glob_iterator = Path(folder_path).glob('*')
    return {
        "event_name": "upload_videos_from_folder_to_project",
        "properties": {
            "Video Count": sum(1 for _ in glob_iterator),
        }
    }


def export_annotation(*args, **kwargs):
    dataset_format = kwargs.get("dataset_format", None)
    if not dataset_format:
        dataset_format = args[2]
    project_type = kwargs.get("project_type", None)
    if not project_type:
        project_type = args[4:5]
        if not project_type:
            project_type = 'Vector'
        else:
            project_type = args[4]
    task = kwargs.get("task", None)
    if not task:
        task = args[5:6]
        if not task:
            task = "object_detection"
        else:
            task = args[5]
    return {
        "event_name": "export_annotation",
        "properties":
            {
                "Format": dataset_format,
                "Project Type": project_type,
                "Task": task
            }
    }


def import_annotation(*args, **kwargs):
    dataset_format = kwargs.get("dataset_format", None)
    if not dataset_format:
        dataset_format = args[2]
    project_type = kwargs.get("project_type", None)
    if not project_type:
        project_type = args[4:5]
        if not project_type:
            project_type = 'Vector'
        else:
            project_type = args[4]
    task = kwargs.get("task", None)
    if not task:
        task = args[5:6]
        if not task:
            task = "object_detection"
        else:
            task = args[5]
    return {
        "event_name": "import_annotation",
        "properties":
            {
                "Format": dataset_format,
                "Project Type": project_type,
                "Task": task
            }
    }


def move_images(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    image_names = kwargs.get("image_names", False)
    if image_names == False:
        image_names = args[0]
        if image_names == None:
            from superannotate.db.images import search_images as sa_search_images
            image_names = sa_search_images(project)
    return {
        "event_name": "move_images",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Image Count":
                    len(image_names),
                "Copy Annotations":
                    bool(args[3:4] or ("include_annotations" in kwargs)),
                "Copy Annotation Status":
                    bool(args[4:5] or ("copy_annotation_status" in kwargs)),
                "Copy Pin":
                    bool(args[5:6] or ("copy_pin" in kwargs)),
            }
    }


def copy_images(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    image_names = kwargs.get("image_names", False)
    if image_names == False:
        image_names = args[1]
        if image_names == None:
            from superannotate.db.images import search_images as sa_search_images
            image_names = sa_search_images(project)
    return {
        "event_name": "copy_images",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Image Count":
                    len(image_names),
                "Copy Annotations":
                    bool(args[3:4] or ("include_annotations" in kwargs)),
                "Copy Annotation Status":
                    bool(args[4:5] or ("copy_annotation_status" in kwargs)),
                "Copy Pin":
                    bool(args[5:6] or ("copy_pin" in kwargs)),
            }
    }


def consensus(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    folder_names = kwargs.get("folder_names", None)
    if not folder_names:
        folder_names = args[1]
    image_list = kwargs.get("image_list", "empty")
    if image_list == "empty":
        image_list = args[3:4]
        if len(image_list) and image_list[0] == None:
            from superannotate.db.images import search_images as sa_search_images
            image_list = sa_search_images(project)
        else:
            image_list = image_list[0]
    annot_type = kwargs.get("annot_type", "empty")
    if annot_type == 'empty':
        annot_type = args[4:5]
        if not annot_type:
            annot_type = "bbox"
        else:
            annot_type = args[4]

    show_plots = kwargs.get("show_plots", "empty")
    if show_plots == "empty":
        show_plots = args[5:6]
        if not show_plots:
            show_plots = False
        else:
            show_plots = args[5]
    return {
        "event_name": "consensus",
        "properties":
            {
                "Folder Count": len(folder_names),
                "Image Count": len(image_list),
                "Annotation Type": annot_type,
                "Plot": show_plots
            }
    }


def benchmark(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    folder_names = kwargs.get("folder_names", None)
    if not folder_names:
        folder_names = args[2]
    image_list = kwargs.get("image_list", "empty")
    if image_list == "empty":
        image_list = args[4:5]
        if len(image_list) and image_list[0] == None:
            from superannotate.db.images import search_images as sa_search_images
            image_list = sa_search_images(project)
        else:
            image_list = image_list[0]
    annot_type = kwargs.get("annot_type", "empty")
    if annot_type == 'empty':
        annot_type = args[5:6]
        if not annot_type:
            annot_type = "bbox"
        else:
            annot_type = args[5]

    show_plots = kwargs.get("show_plots", "empty")
    if show_plots == "empty":
        show_plots = args[6:7]
        if not show_plots:
            show_plots = False
        else:
            show_plots = args[6]
    return {
        "event_name": "benchmark",
        "properties":
            {
                "Folder Count": len(folder_names),
                "Image Count": len(image_list),
                "Annotation Type": annot_type,
                "Plot": show_plots
            }
    }


def upload_annotations_from_folder_to_project(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    from superannotate.db.projects import get_project_metadata as sa_get_project_metadata
    project_name = get_project_name(project)
    project_metadata = sa_get_project_metadata(project_name)

    folder_path = kwargs.get("folder_path", None)
    if not folder_path:
        folder_path = args[1]
    from pathlib import Path
    glob_iterator = Path(folder_path).glob('*.json')
    return {
        "event_name": "upload_annotations_from_folder_to_project",
        "properties":
            {
                "Annotation Count": sum(1 for _ in glob_iterator),
                "Project Type": project_metadata['type'],
                "From S3": bool(args[2:3] or ("from_s3_bucket" in kwargs))
            }
    }


def upload_preannotations_from_folder_to_project(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    from superannotate.db.projects import get_project_metadata as sa_get_project_metadata
    project_name = get_project_name(project)
    project_metadata = sa_get_project_metadata(project_name)
    folder_path = kwargs.get("folder_path", None)
    if not folder_path:
        folder_path = args[1]
    from pathlib import Path
    glob_iterator = Path(folder_path).glob('*.json')
    return {
        "event_name": "upload_preannotations_from_folder_to_project",
        "properties":
            {
                "Annotation Count": sum(1 for _ in glob_iterator),
                "Project Type": project_metadata['type'],
                "From S3": bool(args[2:3] or ("from_s3_bucket" in kwargs))
            }
    }


def upload_images_from_folder_to_project(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    folder_path = kwargs.get("folder_path", None)
    if not folder_path:
        folder_path = args[1]

    recursive_subfolders = kwargs.get("recursive_subfolders", None)
    if not recursive_subfolders:
        recursive_subfolders = args[6:7]
        if recursive_subfolders:
            recursive_subfolders = recursive_subfolders[0]
        else:
            recursive_subfolders = False

    from ...common import DEFAULT_IMAGE_EXTENSIONS
    extensions = DEFAULT_IMAGE_EXTENSIONS
    extensions = kwargs.get("extensions", None)
    if not extensions:
        extensions = args[2:3]
        if extensions:
            extensions = extensions[0]
        else:
            extensions = DEFAULT_IMAGE_EXTENSIONS

    from ...common import DEFAULT_FILE_EXCLUDE_PATTERNS
    exclude_file_patterns = DEFAULT_FILE_EXCLUDE_PATTERNS
    exclude_file_patterns = kwargs.get("exclude_file_patterns", None)
    if not exclude_file_patterns:
        exclude_file_patterns = args[5:6]
        if exclude_file_patterns:
            exclude_file_patterns = exclude_file_patterns[0]
        else:
            exclude_file_patterns = DEFAULT_FILE_EXCLUDE_PATTERNS

    from pathlib import Path
    import os

    paths = []
    for extension in extensions:
        if not recursive_subfolders:
            paths += list(Path(folder_path).glob(f'*.{extension.lower()}'))
            if os.name != "nt":
                paths += list(Path(folder_path).glob(f'*.{extension.upper()}'))
        else:
            paths += list(Path(folder_path).rglob(f'*.{extension.lower()}'))
            if os.name != "nt":
                paths += list(Path(folder_path).rglob(f'*.{extension.upper()}'))

    filtered_paths = []
    for path in paths:
        not_in_exclude_list = [
            x not in Path(path).name for x in exclude_file_patterns
        ]
        if all(not_in_exclude_list):
            filtered_paths.append(path)

    return {
        "event_name": "upload_images_from_folder_to_project",
        "properties":
            {
                "Image Count":
                    len(filtered_paths),
                "Custom Extentions":
                    bool(args[2:3] or kwargs.get("extensions", None)),
                "Annotation Status":
                    bool(args[3:4] or kwargs.get("annotation_status", None)),
                "From S3":
                    bool(args[4:5] or kwargs.get("from_s3_bucket", None)),
                "Custom Exclude Patters":
                    bool(
                        args[5:6] or kwargs.get("exclude_file_patterns", None)
                    )
            }
    }


def upload_images_from_s3_bucket_to_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "upload_images_from_s3_bucket_to_project",
        "properties": {
            "project_name": get_project_name(project)
        }
    }


def prepare_export(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "prepare_export",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Folder Count":
                    bool(args[1:2] or kwargs.get("folder_names", None)),
                "Annotation Statuses":
                    bool(args[2:3] or kwargs.get("annotation_statuses", None)),
                "Include Fuse":
                    bool(args[3:4] or kwargs.get("include_fuse", None)),
                "Only Pinned":
                    bool(args[4:5] or kwargs.get("only_pinned", None)),
            }
    }


def download_export(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "download_export",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "to_s3_bucket":
                    bool(args[4:5] or kwargs.get("to_s3_bucket", None)),
            }
    }


def dicom_to_rgb_sequence(*args, **kwargs):
    return {"event_name": "dicom_to_rgb_sequence", "properties": {}}


def coco_split_dataset(*args, **kwargs):
    ratio_list = kwargs.get("ratio_list", None)
    if not ratio_list:
        ratio_list = args[4]
    return {
        "event_name": "coco_split_dataset",
        "properties": {
            "ratio_list": str(ratio_list)
        }
    }


def run_training(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0][0]
    task = kwargs.get("task", None)
    if not task:
        task = args[4]
    hyperparameters = kwargs.get("hyperparameters", None)
    if not hyperparameters:
        hyperparameters = args[4]
    from superannotate.db.projects import get_project_metadata as sa_get_project_metadata
    project_name = get_project_name(project)
    project_metadata = sa_get_project_metadata(project_name)
    log = kwargs.get("log", "empty")
    if log == "empty":
        log = args[6:7]
        if not log:
            log = False
        else:
            log = args[6]
    return {
        "event_name": "run_training",
        "properties":
            {
                "Project Type": project_metadata['type'],
                "Task": task,
                "Learning Rate": hyperparameters['base_lr'],
                "Batch Size": hyperparameters['images_per_batch'],
                "Log": log
            }
    }


def assign_images(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    image_names = kwargs.get("image_names", None)
    if not image_names:
        image_names = args[1]
    user = kwargs.get("user", None)
    if not user:
        user = args[2]
    from superannotate.db.users import get_team_contributor_metadata
    res = get_team_contributor_metadata(user)
    user_role = "ADMIN"
    if res['user_role'] == 3:
        user_role = 'ANNOTATOR'
    if res['user_role'] == 4:
        user_role = 'QA'
    from superannotate.db.project_api import get_project_and_folder_metadata
    project, folder = get_project_and_folder_metadata(project)
    is_root = True
    if folder:
        is_root = False
    return {
        "event_name": "assign_images",
        "properties":
            {
                "project_name": get_project_name(project),
                "Assign Folder": is_root,
                "Image Count": len(image_names),
                "User Role": user_role,
            }
    }


def move_image(*args, **kwargs):
    project = kwargs.get("source_project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "move_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Move Annotations":
                    bool(args[3:4] or ("include_annotations" in kwargs)),
                "Move Annotation Status":
                    bool(args[4:5] or ("copy_annotation_status" in kwargs)),
                "Move Pin":
                    bool(args[5:6] or ("copy_pin" in kwargs)),
            }
    }


def pin_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "pin_image",
        "properties":
            {
                "project_name": get_project_name(project),
                "Pin": bool(args[2:3] or ("pin" in kwargs)),
            }
    }


def create_fuse_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    project_type = kwargs.get("project_type", None)
    if not project_type:
        project_type = args[2]
    return {
        "event_name": "create_fuse_image",
        "properties":
            {
                "project_name": get_project_name(project),
                "Project Type": project_type,
                "Overlay": bool(args[4:5] or ("output_overlay" in kwargs)),
            }
    }


def set_image_annotation_status(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "set_image_annotation_status",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Annotation Status":
                    bool(args[2:3] or ("annotation_status" in kwargs)),
            }
    }


def add_annotation_bbox_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_bbox_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def add_annotation_polygon_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_polygon_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def add_annotation_polyline_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_polyline_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def add_annotation_point_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_point_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def add_annotation_ellipse_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_ellipse_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def add_annotation_template_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_template_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[5:6] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[6:7] or ("error" in kwargs)),
            }
    }


def add_annotation_cuboid_to_image(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "add_annotation_cuboid_to_image",
        "properties":
            {
                "project_name":
                    get_project_name(project),
                "Attributes":
                    bool(
                        args[4:5] or ("annotation_class_attributes" in kwargs)
                    ),
                "Error":
                    bool(args[5:6] or ("error" in kwargs)),
            }
    }


def create_annotation_class(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]

    return {
        "event_name": "create_annotation_class",
        "properties":
            {
                "project_name": get_project_name(project),
                "Attributes": bool(args[3:4] or ("attribute_groups" in kwargs)),
            }
    }


def create_annotation_classes_from_classes_json(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "create_annotation_classes_from_classes_json",
        "properties":
            {
                "project_name": get_project_name(project),
                "From S3": bool(args[2:3] or ("from_s3_bucket" in kwargs)),
            }
    }


def class_distribution(*args, **kwargs):
    return {
        "event_name": "class_distribution",
        "properties": {
            "Plot": bool(args[2:3] or ("visualize" in kwargs)),
        }
    }


def attribute_distribution(*args, **kwargs):
    return {
        "event_name": "attribute_distribution",
        "properties": {
            "Plot": bool(args[2:3] or ("visualize" in kwargs)),
        }
    }


def share_project(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    user_role = kwargs.get("user_role", None)
    if not user_role:
        user_role = args[2]
    return {
        "event_name": "share_project",
        "properties":
            {
                "project_name": get_project_name(project),
                "User Role": user_role
            }
    }


def set_project_default_image_quality_in_editor(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    image_quality_in_editor = kwargs.get("image_quality_in_editor", None)
    if not image_quality_in_editor:
        image_quality_in_editor = args[1]
    return {
        "event_name": "set_project_default_image_quality_in_editor",
        "properties":
            {
                "project_name": get_project_name(project),
                "Image Quality": image_quality_in_editor
            }
    }


def get_exports(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "get_exports",
        "properties":
            {
                "project_name": get_project_name(project),
                "Metadata": bool(args[1:2] or ("return_metadata" in kwargs)),
            }
    }


def search_folders(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    return {
        "event_name": "search_folders",
        "properties":
            {
                "project_name": get_project_name(project),
                "Metadata": bool(args[2:3] or ("return_metadata" in kwargs)),
            }
    }


def filter_images_by_tags(*args, **kwargs):
    return {
        "event_name": "filter_images_by_tags",
        "properties":
            {
                "Include": bool(args[1:2] or ("include" in kwargs)),
                "Exclude": bool(args[2:3] or ("exclude" in kwargs))
            }
    }


def filter_images_by_comments(*args, **kwargs):
    return {
        "event_name": "filter_images_by_comments",
        "properties":
            {
                "Include Unresolved Comments":
                    bool(
                        args[1:2] or ("include_unresolved_comments" in kwargs)
                    ),
                "Include Resolved Comments":
                    bool(args[2:3] or ("include_resolved_comments" in kwargs))
            }
    }


def filter_annotation_instances(*args, **kwargs):
    return {
        "event_name": "filter_annotation_instances",
        "properties":
            {
                "Include": bool(args[1:2] or ("include" in kwargs)),
                "Exclude": bool(args[2:3] or ("exclude" in kwargs))
            }
    }


def aggregate_annotations_as_df(*args, **kwargs):

    folder_names = kwargs.get("folder_names", "empty")
    if folder_names == "empty":
        folder_names = args[5:6]
        if folder_names != None:
            folder_names = len(folder_names)
    return {
        "event_name": "aggregate_annotations_as_df",
        "properties": {
            "Folder Count": folder_names,
        }
    }


def delete_folders(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    folder_names = kwargs.get("folder_names", None)
    if not folder_names:
        folder_names = args[1]
    return {
        "event_name": "delete_folders",
        "properties":
            {
                "project_name": get_project_name(project),
                "Folder Count": len(folder_names),
            }
    }


def delete_images(*args, **kwargs):
    project = kwargs.get("project", None)
    if not project:
        project = args[0]
    image_names = kwargs.get("image_names", None)
    if not image_names:
        image_names = args[1]
    return {
        "event_name": "delete_images",
        "properties":
            {
                "project_name": get_project_name(project),
                "Image Count": len(image_names),
            }
    }



'''
'''
import numpy as np
from ..baseStrategy import baseStrategy

from ....common import write_to_json


class DataLoopStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        classes = self.conversion_algorithm(
            self.export_root, self.task, self.output_dir
        )
        sa_classes = self._create_sa_classes(classes)
        (self.output_dir / 'classes').mkdir(exist_ok=True)
        write_to_json(self.output_dir / 'classes' / 'classes.json', sa_classes)

    def _create_sa_classes(self, classes):
        sa_classes_loader = []
        for key, value in classes.items():
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            sa_classes = {
                'name': key,
                'color': hexcolor,
                'attribute_groups': []
            }
            attribute_groups = []
            for attr_group_key, attr_group in value['attribute_group'].items():
                attr_loader = {
                    'name': attr_group_key,
                    'is_multiselect': attr_group['is_multiselect'],
                    'attributes': []
                }
                for attr in set(attr_group['attributes']):
                    attr_loader['attributes'].append({'name': attr})
                if attr_loader:
                    attribute_groups.append(attr_loader)

            sa_classes['attribute_groups'] = attribute_groups

            sa_classes_loader.append(sa_classes)
        return sa_classes_loader



import numpy as np


def _update_classes_dict(classes, new_class, new_attributes):
    if new_class not in classes.keys():
        classes[new_class] = {'attribute_group': {}}
        classes[new_class]['attribute_group']['converted_attributes'] = {}
        classes[new_class]['attribute_group']['converted_attributes'][
            'is_multiselect'] = 1
        classes[new_class]['attribute_group']['converted_attributes'][
            'attributes'] = new_attributes
    else:
        classes[new_class]['attribute_group']['converted_attributes'][
            'attributes'] += new_attributes
    return classes


def _create_attributes_list(dl_attributes):
    attributes = []
    for attribute in dl_attributes:
        attr = {'name': attribute, 'groupName': 'converted_attributes'}
        attributes.append(attr)
    return attributes



from pathlib import Path

import superannotate as sa
from ..common import upload_project


def test_supervisely_convert_vector(tmpdir):
    project_name = "supervisely_test_vector_basic"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'Supervisely' / 'input' / 'toSuperAnnotate' / 'vector'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'Supervisely', '', 'Vector', 'vector_annotation'
    )

    description = 'supervisely vector annotation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_supervisely_convert_object(tmpdir):
    project_name = "supervisely_test_object"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'Supervisely' / 'input' / 'toSuperAnnotate' / 'vector'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'Supervisely', '', 'Vector', 'object_detection'
    )

    description = 'supervisely object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_supervisely_convert_instance(tmpdir):
    project_name = "supervisely_test_vector_instance"
    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'Supervisely' / 'input' / 'toSuperAnnotate' / 'vector'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'Supervisely', '', 'Vector', 'instance_segmentation'
    )

    description = 'supervisely instance segmentation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_supervisely_convert_keypoint(tmpdir):
    project_name = "supervisely_test_keypoint"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'Supervisely' / 'input' / 'toSuperAnnotate' / 'keypoints'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'Supervisely', '', 'Vector', 'keypoint_detection'
    )

    description = 'supervisely keypoint'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_supervisely_convert_instance_pixel(tmpdir):
    project_name = "supervisely_test_instance_pixel"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'Supervisely' / 'input' / 'toSuperAnnotate' / 'instance'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'Supervisely', '', 'Pixel', 'instance_segmentation'
    )

    description = 'supervisely instance segmentation'
    ptype = 'Pixel'
    upload_project(out_dir, project_name, description, ptype)



'''
Dataloop to SA conversion method
'''
import logging
import threading
import json

from .dataloop_helper import (_update_classes_dict, _create_attributes_list)

from ..sa_json_helper import (
    _create_vector_instance, _create_comment, _create_sa_json
)

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def dataloop_to_sa(input_dir, task, output_dir):
    classes = {}
    json_data = list(input_dir.glob('*.json'))
    if task == 'object_detection':
        instance_types = ['box']
    elif task == 'instance_segmentation':
        instance_types = ['segment']
    elif task == 'vector_annotation':
        instance_types = ['point', 'box', 'ellipse', 'segment']

    tags_type = 'class'
    comment_type = 'note'

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_data), images_converted, images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in json_data:
        dl_data = json.load(open(json_file))

        sa_metadata = {}
        if 'itemMetadata' in dl_data and 'system' in dl_data['itemMetadata']:
            temp = dl_data['itemMetadata']['system']
            sa_metadata['name'] = temp['originalname']
            sa_metadata['width'] = temp['width']
            sa_metadata['height'] = temp['height']

        sa_instances = []
        sa_tags = []
        sa_comments = []

        for ann in dl_data['annotations']:
            if ann['type'] in instance_types:
                classes = _update_classes_dict(
                    classes, ann['label'], ann['attributes']
                )

            attributes = _create_attributes_list(ann['attributes'])

            if ann['type'] in instance_types:
                if ann['type'] == 'segment' and len(ann['coordinates']) == 1:
                    points = []
                    for sub_list in ann['coordinates']:
                        for sub_dict in sub_list:
                            points.append(sub_dict['x'])
                            points.append(sub_dict['y'])
                    instance_type = 'polygon'
                elif ann['type'] == 'box':
                    points = (
                        ann['coordinates'][0]['x'], ann['coordinates'][0]['y'],
                        ann['coordinates'][1]['x'], ann['coordinates'][1]['y']
                    )
                    instance_type = 'bbox'
                elif ann['type'] == 'ellipse':
                    points = (
                        ann['coordinates']['center']['x'],
                        ann['coordinates']['center']['y'],
                        ann['coordinates']['rx'], ann['coordinates']['ry'],
                        ann['coordinates']['angle']
                    )
                    instance_type = 'ellipse'
                elif ann['type'] == 'point':
                    points = (ann['coordinates']['x'], ann['coordinates']['y'])
                    instance_type = 'point'
                sa_obj = _create_vector_instance(
                    instance_type, points, {}, attributes, ann['label']
                )
                sa_instances.append(sa_obj)
            elif ann['type'] == comment_type:
                points = (
                    ann['coordinates']['box'][0]['x'],
                    ann['coordinates']['box'][0]['y']
                )
                comments = []
                for note in ann['coordinates']['note']['messages']:
                    comments.append(
                        {
                            'text': note['body'],
                            'email': note['creator']
                        }
                    )
                    sa_comment = _create_comment(points, comments)
                sa_comments.append(sa_comment)
            elif ann['type'] == tags_type:
                sa_tags.append(ann['label'])

        if 'name' in sa_metadata:
            file_name = '%s___objects.json' % sa_metadata['name']
        else:
            file_name = '%s___objects.json' % dl_data['filename'][1:]

        images_converted.append(file_name.replace('___objects.json ', ''))
        json_template = _create_sa_json(
            sa_instances, sa_metadata, sa_tags, sa_comments
        )
        write_to_json(output_dir / file_name, json_template)
    finish_event.set()
    tqdm_thread.join()
    return classes



'''
Sagemaker to SA conversion method
'''
import logging
import threading
import json
from pathlib import Path

from ..sa_json_helper import (_create_vector_instance, _create_sa_json)

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def sagemaker_object_detection_to_sa_vector(data_path, main_key, output_dir):
    dataset_manifest = []
    try:
        img_map_file = open(data_path / 'output.manifest')
    except Exception as e:
        raise Exception("'output.manifest' file doesn't exist")

    for line in img_map_file:
        dataset_manifest.append(json.loads(line))

    json_list = list(data_path.glob('*.json'))
    classes_ids = {}
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_list), images_converted, images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in json_list:
        data_json = json.load(open(json_file))
        for img in data_json:
            if 'consolidatedAnnotation' not in img.keys():
                logger.warning('Wrong json files')
                raise Exception

            manifest = dataset_manifest[int(img['datasetObjectId'])]
            file_name = '%s___objects.json' % Path(manifest['source-ref']).name

            classes = img['consolidatedAnnotation']['content'][
                main_key + '-metadata']['class-map']
            for key, value in classes.items():
                if key not in classes_ids.keys():
                    classes_ids[key] = value

            image_size = img['consolidatedAnnotation']['content'][main_key][
                'image_size'][0]

            sa_metadata = {
                'name': Path(manifest['source-ref']).name,
                'width': image_size['width'],
                'height': image_size['height']
            }

            annotations = img['consolidatedAnnotation']['content'][main_key][
                'annotations']
            sa_instances = []
            for annotation in annotations:
                points = (
                    annotation['left'], annotation['top'],
                    annotation['left'] + annotation['width'],
                    annotation['top'] + annotation['height']
                )
                sa_obj = _create_vector_instance(
                    'bbox', points, {}, [], classes[str(annotation['class_id'])]
                )
                sa_instances.append(sa_obj.copy())
            images_converted.append(Path(manifest['source-ref']).name)
            sa_json = _create_sa_json(sa_instances, sa_metadata)
            write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes_ids



import functools
import json
import logging
import os
import sys
import time
from pathlib import Path

import numpy as np
from PIL import Image
from tqdm import tqdm

logger = logging.getLogger("superannotate-python-sdk")

DEFAULT_IMAGE_EXTENSIONS = ("jpg", "jpeg", "png", "tif", "tiff", "webp", "bmp")
DEFAULT_FILE_EXCLUDE_PATTERNS = ("___save.png", "___fuse.png")

DEFAULT_VIDEO_EXTENSIONS = ("mp4", "avi", "mov", "webm", "flv", "mpg", "ogg")

SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES = set('/\\:*?"<>|')

_PROJECT_TYPES = {"Vector": 1, "Pixel": 2}

_ANNOTATION_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "QualityCheck": 3,
    "Returned": 4,
    "Completed": 5,
    "Skipped": 6
}

_UPLOAD_STATES_STR_TO_CODES = {"Initial": 1, "Basic": 2, "External": 3}
_UPLOAD_STATES_CODES_TO_STR = {1: "Initial", 2: "Basic", 3: "External"}

_USER_ROLES = {"Admin": 2, "Annotator": 3, "QA": 4, "Customer": 5, "Viewer": 6}
_AVAILABLE_SEGMENTATION_MODELS = ['autonomous', 'generic']
_MODEL_TRAINING_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "Completed": 3,
    "FailedBeforeEvaluation": 4,
    "FailedAfterEvaluation": 5,
    "FailedAfterEvaluationWithSavedModel": 6
}

_PREDICTION_SEGMENTATION_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "Completed": 3,
    "Failed": 4
}

_MODEL_TRAINING_TASKS = {
    "Instance Segmentation for Pixel Projects": "instance_segmentation_pixel",
    "Instance Segmentation for Vector Projects": "instance_segmentation_vector",
    "Keypoint Detection for Vector Projects": "keypoint_detection_vector",
    "Object Detection for Vector Projects": "object_detection_vector",
    "Semantic Segmentation for Pixel Projects": "semantic_segmentation_pixel"
}


def prediction_segmentation_status_from_str_to_int(status):
    return _PREDICTION_SEGMENTATION_STATUSES[status]


def prediction_segmentation_status_from_int_to_str(status):
    for idx, item in _PREDICTION_SEGMENTATION_STATUSES.items():
        if item == status:
            return idx
    raise RuntimeError("NA segmentation/prediction status")


def model_training_status_int_to_str(project_status):
    for item in _MODEL_TRAINING_STATUSES:
        if _MODEL_TRAINING_STATUSES[item] == project_status:
            return item
    raise RuntimeError("NA training status")


def model_training_status_str_to_int(project_status):
    return _MODEL_TRAINING_STATUSES[project_status]


def image_path_to_annotation_paths(image_path, project_type):
    image_path = Path(image_path)
    if project_type == "Vector":
        return (
            image_path.parent /
            get_annotation_json_name(image_path.name, project_type),
        )
    return (
        image_path.parent /
        get_annotation_json_name(image_path.name, project_type),
        image_path.parent / get_annotation_png_name(image_path.name)
    )


def project_type_str_to_int(project_type):
    return _PROJECT_TYPES[project_type]


def project_type_int_to_str(project_type):
    """Converts metadata project_type int value to a string

    :param project_type: int in project metadata's 'type' key
    :type project_type: int

    :return: 'Vector' or 'Pixel'
    :rtype: str
    """
    for k, v in _PROJECT_TYPES.items():
        if v == project_type:
            return k
    raise RuntimeError("NA Project type")


def user_role_str_to_int(user_role):
    return _USER_ROLES[user_role]


def user_role_int_to_str(user_role):
    for k, v in _USER_ROLES.items():
        if v == user_role:
            return k
    return None


def annotation_status_str_to_int(annotation_status):
    return _ANNOTATION_STATUSES[annotation_status]


def upload_state_str_to_int(upload_state):
    return _UPLOAD_STATES_STR_TO_CODES[upload_state]


def upload_state_int_to_str(upload_state):
    return _UPLOAD_STATES_CODES_TO_STR[upload_state]


def annotation_status_int_to_str(annotation_status):
    """Converts metadata annotation_status int value to a string

    :param annotation_status: int in image metadata's 'annotation_status' key
    :type annotation_status: int

    :return: One of 'NotStarted' 'InProgress' 'QualityCheck' 'Returned' 'Completed' 'Skipped'
    :rtype: str
    """

    for k, v in _ANNOTATION_STATUSES.items():
        if v == annotation_status:
            return k
    return None


def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(
                    '{} received both {} and {}'.format(func_name, alias, new)
                )
            logger.warning(
                '%s is deprecated; use %s in %s', alias, new, func_name
            )
            kwargs[new] = kwargs.pop(alias)


def hex_to_rgb(hex_string):
    """Converts HEX values to RGB values
    """
    h = hex_string.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(rgb_tuple):
    """Converts RGB values to HEX values
    """
    return '#%02x%02x%02x' % rgb_tuple


def blue_color_generator(n, hex_values=True):
    """ Blue colors generator for SuperAnnotate blue mask.
    """
    hex_colors = []
    for i in range(n + 1):
        int_color = i * 15
        bgr_color = np.array(
            [
                int_color & 255, (int_color >> 8) & 255,
                (int_color >> 16) & 255, 255
            ],
            dtype=np.uint8
        )
        hex_color = '#' + "{:02x}".format(
            bgr_color[2]
        ) + "{:02x}".format(bgr_color[1], ) + "{:02x}".format(bgr_color[0])
        if hex_values:
            hex_colors.append(hex_color)
        else:
            hex_colors.append(hex_to_rgb(hex_color))
    return hex_colors[1:]


def id2rgb(id_map):
    if isinstance(id_map, np.ndarray):
        id_map_copy = id_map.copy()
        rgb_shape = tuple(list(id_map.shape) + [3])
        rgb_map = np.zeros(rgb_shape, dtype=np.uint8)
        for i in range(3):
            rgb_map[..., i] = id_map_copy % 256
            id_map_copy //= 256
        return rgb_map
    color = []
    for _ in range(3):
        color.append(id_map % 256)
        id_map //= 256
    return color


def save_desktop_format(output_dir, classes, files_dict):
    cat_id_map = {}
    new_classes = []
    for idx, class_ in enumerate(classes):
        cat_id_map[class_['name']] = idx + 2
        class_['id'] = idx + 2
        new_classes.append(class_)
    with open(output_dir.joinpath('classes.json'), 'w') as fw:
        json.dump(new_classes, fw)

    meta = {
        "type": "meta",
        "name": "lastAction",
        "timestamp": int(round(time.time() * 1000))
    }
    new_json = {}
    files_path = []
    (output_dir / 'images' / 'thumb').mkdir()
    for file_name, json_data in files_dict.items():
        file_name = file_name.replace('___objects.json', '')
        if not (output_dir / 'images' / file_name).exists():
            continue

        for js_data in json_data:
            if 'className' in js_data:
                js_data['classId'] = cat_id_map[js_data['className']]
        json_data.append(meta)
        new_json[file_name] = json_data

        files_path.append(
            {
                'srcPath':
                    str(output_dir.resolve() / file_name),
                'name':
                    file_name,
                'imagePath':
                    str(output_dir.resolve() / file_name),
                'thumbPath':
                    str(
                        output_dir.resolve() / 'images' / 'thumb' /
                        ('thmb_' + file_name + '.jpg')
                    ),
                'valid':
                    True
            }
        )

        img = Image.open(output_dir / 'images' / file_name)
        img.thumbnail((168, 120), Image.ANTIALIAS)
        img.save(
            output_dir / 'images' / 'thumb' / ('thmb_' + file_name + '.jpg')
        )

    with open(output_dir / 'images' / 'images.sa', 'w') as fw:
        fw.write(json.dumps(files_path))

    with open(output_dir.joinpath('annotations.json'), 'w') as fw:
        json.dump(new_json, fw)

    with open(output_dir / 'config.json', 'w') as fw:
        json.dump({"pathSeparator": os.sep, "os": sys.platform}, fw)


def save_web_format(output_dir, classes, files_dict):
    for key, value in files_dict.items():
        with open(output_dir.joinpath(key), 'w') as fw:
            json.dump(value, fw, indent=2)

    with open(output_dir.joinpath('classes', 'classes.json'), 'w') as fw:
        json.dump(classes, fw)


def dump_output(output_dir, platform, classes, files_dict):
    if platform == 'Web':
        save_web_format(output_dir, classes, files_dict)
    else:
        save_desktop_format(output_dir, classes, files_dict)


def write_to_json(output_path, json_data):
    with open(output_path, 'w') as fw:
        json.dump(json_data, fw, indent=2)


MAX_IMAGE_SIZE = 100 * 1024 * 1024  # 100 MB limit
MAX_IMAGE_RESOLUTION = {
    "Vector": 100_000_000,
    "Pixel": 4_000_000
}  # Resolution limit


def tqdm_converter(
    total_num, images_converted, images_not_converted, finish_event
):
    with tqdm(total=total_num) as pbar:
        while True:
            finished = finish_event.wait(5)
            if not finished:
                sum_all = len(images_converted) + len(images_not_converted)
                pbar.update(sum_all - pbar.n)
            else:
                pbar.update(total_num - pbar.n)
                break


def get_annotation_json_name(image_name, project_type):
    if project_type == "Vector":
        return image_name + "___objects.json"
    else:
        return image_name + "___pixel.json"


def get_annotation_png_name(image_name):
    return image_name + "___save.png"



from pathlib import Path
import superannotate as sa
from ..common import upload_project


def test_googlecloud_convert_web(tmpdir):
    project_name = "googlcloud_object"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "GoogleCloud" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "GoogleCloud", "image_object_detection", "Vector",
        "object_detection"
    )

    description = 'googlecloud object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)



from ..db.images import get_image_metadata
from ..common import _PREDICTION_SEGMENTATION_STATUSES
from .defaults import PLOTTABLE_METRICS
from ast import literal_eval
import pandas as pd
import time
import os


def metric_is_plottable(key):
    if key in PLOTTABLE_METRICS or 'mIoU' in key or 'mAP' in key or key == 'iteration':
        return True
    return False


def reformat_metrics_json(data, name):
    continuous_metrics = []
    per_evaluation_metrics = []
    for item in data:
        for key in item:
            try:
                item[key] = literal_eval(item[key])
            except Exception as e:
                pass
        if 'data_time' in item and item['eta_seconds'] != 0:
            continuous_metrics += [item]
        else:
            per_evaluation_metrics += [item]
    continuous_metrics_df = pd.DataFrame.from_dict(continuous_metrics)
    per_evaluation_metrics_df = pd.DataFrame.from_dict(per_evaluation_metrics)
    continuous_metrics_df = drop_non_plotable_cols(continuous_metrics_df)
    per_evaluation_metrics_df = drop_non_plotable_cols(
        per_evaluation_metrics_df
    )
    continuous_metrics_df['model'] = name
    per_evaluation_metrics_df['model'] = name
    if 'total_loss' in per_evaluation_metrics_df:
        per_evaluation_metrics_df = per_evaluation_metrics_df.drop(
            columns='total_loss'
        )

    per_evaluation_metrics_df = per_evaluation_metrics_df.dropna(axis='rows')
    return continuous_metrics_df, per_evaluation_metrics_df


def drop_non_plotable_cols(df):
    for column in df:
        if metric_is_plottable(column):
            continue
        df = df.drop(columns=column)
    return df


def make_plotly_specs(num_rows):
    specs = [[{"secondary_y": True}] for x in range(num_rows)]
    return specs


def get_images_prediction_segmentation_status(project, image_names, task):
    metadata = get_image_metadata(project, image_names)
    if isinstance(metadata, dict):
        metadata = [metadata]

    success_names = [x['name'] for x in metadata if x[task] == 'Completed']
    failure_names = [x['name'] for x in metadata if x[task] == 'Failed']
    return success_names, failure_names


def log_process(
    project, image_name_set, total_image_count, status_key, task, logger
):
    num_complete = 0
    succeded_imgs = []
    failed_imgs = []
    while image_name_set:
        succeded_imgs_batch, failed_imgs_batch = get_images_prediction_segmentation_status(
            project, list(image_name_set), status_key
        )
        complete_images = succeded_imgs_batch + failed_imgs_batch
        succeded_imgs += succeded_imgs_batch
        failed_imgs += failed_imgs_batch
        num_complete += len(complete_images)
        logger.info(
            f"{task} complete on {num_complete} / {total_image_count} images"
        )
        image_name_set = image_name_set.symmetric_difference(
            set(complete_images)
        )
        time.sleep(5)
    return (succeded_imgs, failed_imgs)



import math
from tqdm import tqdm
import threading
import io
import time
from pathlib import Path
from .. import common
import logging
import uuid
from PIL import Image, ImageOps
import json
from ..api import API
from ..exceptions import SABaseException, SAImageSizeTooLarge, SANonExistingProjectNameException
import datetime
import boto3

_api = API.get_instance()
logger = logging.getLogger("superannotate-python-sdk")


# Yield successive n-sized
# chunks from l.
def divide_chunks(l, n):

    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]


def get_project_folder_string(project):
    if isinstance(project, dict):
        return project['name']
    elif isinstance(project, tuple):
        project, folder = project
        project_name = project['name']
        if folder:
            return project_name + '/' + folder['name']
        return project_name
    elif isinstance(project, str):
        return project


def __move_images(
    source_project, source_folder_id, destination_folder_id, image_names
):
    """Move images in bulk between folders in a project 

    :param source_project: source project
    :type source_project: dict
    :param source_folder_id: source folder id
    :type source_folder_id: int
    :param image_names: image names. If None, all images from source project will be moved
    :type image: list of str
    :param destination_folder_id: destination folder id
    :type destination_folder_id: int
    :return: tuple of moved images list and skipped images list
    :rtype: tuple of lists
    """

    image_names_lists = divide_chunks(image_names, 1000)
    total_skipped = []
    total_moved = []
    logs = []

    for image_names in image_names_lists:
        response = _api.send_request(
            req_type='POST',
            path='/image/move',
            params={
                "team_id": source_project["team_id"],
                "project_id": source_project["id"]
            },
            json_req={
                "image_names": image_names,
                "destination_folder_id": destination_folder_id,
                "source_folder_id": source_folder_id
            }
        )

        if not response.ok:
            logs.append("Couldn't move images " + response.text)
            total_skipped += image_names
            continue
        res = response.json()
        total_moved += res['done']

    total_skipped = list(set(image_names) - set(total_moved))
    return (total_moved, total_skipped, logs)


def _copy_images_request(
    team_id, project_id, image_names, destination_folder_id, source_folder_id,
    include_annotations, copy_pin
):
    response = _api.send_request(
        req_type='POST',
        path='/images/copy-image-or-folders',
        params={
            "team_id": team_id,
            "project_id": project_id
        },
        json_req={
            "is_folder_copy": False,
            "image_names": image_names,
            "destination_folder_id": destination_folder_id,
            "source_folder_id": source_folder_id,
            "include_annotations": include_annotations,
            "keep_pin_status": copy_pin
        }
    )
    return response


def copy_polling(image_names, source_project, poll_id):
    done_count = 0
    skipped_count = 0
    now_timestamp = datetime.datetime.now().timestamp()
    delta_seconds = len(image_names) * 0.3
    max_timestamp = now_timestamp + delta_seconds
    logs = []
    while True:
        time.sleep(4)
        now_timestamp = datetime.datetime.now().timestamp()
        if (now_timestamp > max_timestamp):
            break
        response = _api.send_request(
            req_type='GET',
            path='/images/copy-image-progress',
            params={
                "team_id": source_project["team_id"],
                "project_id": source_project["id"],
                "poll_id": poll_id
            }
        )
        if not response.ok:
            logs.append("Couldn't copy images " + response.text)
            continue
        res = response.json()
        done_count = int(res['done'])
        skipped_count = int(res['skipped'])
        total_count = int(res['total_count'])
        if (skipped_count + done_count == total_count):
            break
    return (skipped_count, done_count, logs)


def __copy_images(
    source_project, source_folder_id, destination_folder_id, image_names,
    include_annotations, copy_pin
):
    """Copy images in bulk between folders in a project 

    :param source_project: source project
    :type source_project: dict
    :param source_folder_id: source folder id
    :type source_folder_id: int
    :param image_names: image names. If None, all images from source project will be moved
    :type image: list of str
    :param destination_folder_id: destination folder id
    :type destination_folder_id: int
    :param include_annotations: enables annotations copy
    :type include_annotations: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    """

    team_id = source_project["team_id"]
    project_id = source_project["id"]

    image_names_lists = divide_chunks(image_names, 1000)
    total_skipped_count = 0
    total_done_count = 0
    total_skipped_list = []

    logs = []

    for image_names in image_names_lists:
        duplicates = get_duplicate_image_names(
            project_id=project_id,
            team_id=team_id,
            folder_id=destination_folder_id,
            image_paths=image_names
        )
        total_skipped_list += duplicates
        image_names = list(set(image_names) - set(duplicates))
        if not image_names:
            continue

        response = _copy_images_request(
            team_id, project_id, image_names, destination_folder_id,
            source_folder_id, include_annotations, copy_pin
        )
        if not response.ok:
            logs.append("Couldn't copy images " + response.text)
            total_skipped_list += image_names
            continue

        res = response.json()
        poll_id = res['poll_id']
        skipped_count, done_count, polling_logs = copy_polling(
            image_names, source_project, poll_id
        )
        logs += polling_logs
        total_skipped_count += skipped_count
        total_done_count += done_count
    return (total_done_count, total_skipped_list, logs)


def create_empty_annotation(size, image_name):
    return {
        "metadata": {
            'height': size[1],
            'width': size[0],
            'name': image_name
        }
    }


def upload_image_array_to_s3(
    bucket, img_name, img_name_hash, size, orig_image, lores_image, huge_image,
    thumbnail_image, prefix
):
    key = prefix + img_name_hash
    bucket.put_object(Body=orig_image, Key=key)
    bucket.put_object(Body=lores_image, Key=key + '___lores.jpg')
    bucket.put_object(
        Body=huge_image,
        Key=key + '___huge.jpg',
        Metadata={
            'height': str(size[1]),
            'width': str(size[0])
        }
    )
    bucket.put_object(Body=thumbnail_image, Key=key + '___thumb.jpg')
    bucket.put_object(
        Body=json.dumps(create_empty_annotation(size, img_name)),
        Key=key + ".json"
    )
    return key


def get_image_array_to_upload(
    img_name, byte_io_orig, image_quality_in_editor, project_type
):
    if image_quality_in_editor not in ["original", "compressed"]:
        raise SABaseException(0, "NA ImageQuality in get_image_array_to_upload")
    Image.MAX_IMAGE_PIXELS = None
    im = Image.open(byte_io_orig)
    im_format = im.format

    im = ImageOps.exif_transpose(im)

    width, height = im.size

    resolution = width * height
    if resolution > common.MAX_IMAGE_RESOLUTION[project_type]:
        raise SABaseException(
            0, "Image resolution " + str(resolution) +
            " too large. Max supported for " + project_type + " projects is " +
            str(common.MAX_IMAGE_RESOLUTION[project_type])
        )

    if image_quality_in_editor == "original" and im_format in ['JPEG', 'JPG']:
        byte_io_lores = io.BytesIO(byte_io_orig.getbuffer())
    else:
        byte_io_lores = io.BytesIO()
        bg = Image.new('RGBA', im.size, (255, 255, 255))
        im = im.convert("RGBA")
        bg.paste(im, mask=im)
        bg = bg.convert('RGB')
        if image_quality_in_editor == "original":
            bg.save(byte_io_lores, 'JPEG', quality=100, subsampling=0)
        else:
            bg.save(byte_io_lores, 'JPEG', quality=60)
        im = bg

    byte_io_huge = io.BytesIO()
    hsize = int(height * 600.0 / width)
    im.resize((600, hsize), Image.ANTIALIAS).save(byte_io_huge, 'JPEG')

    byte_io_thumbs = io.BytesIO()
    thumbnail_size = (128, 96)
    background = Image.new('RGB', thumbnail_size, "black")
    im.thumbnail(thumbnail_size, Image.ANTIALIAS)
    (w, h) = im.size
    background.paste(
        im, ((thumbnail_size[0] - w) // 2, (thumbnail_size[1] - h) // 2)
    )
    im = background
    im.save(byte_io_thumbs, 'JPEG')

    byte_io_thumbs.seek(0)
    byte_io_lores.seek(0)
    byte_io_huge.seek(0)
    byte_io_orig.seek(0)

    img_name_hash = str(uuid.uuid4()) + Path(img_name).suffix
    return img_name, img_name_hash, (
        width, height
    ), byte_io_orig, byte_io_lores, byte_io_huge, byte_io_thumbs


def __upload_images_to_aws_thread(
    res, img_paths, project, annotation_status, prefix, thread_id, chunksize,
    couldnt_upload, uploaded, tried_upload, image_quality_in_editor,
    from_s3_bucket, project_folder_id
):
    len_img_paths = len(img_paths)
    start_index = thread_id * chunksize
    end_index = start_index + chunksize
    if from_s3_bucket is not None:
        from_session = boto3.Session()
        from_s3 = from_session.resource('s3')
    if start_index >= len_img_paths:
        return
    s3_session = _get_boto_session_by_credentials(res)
    s3_resource = s3_session.resource('s3')
    bucket = s3_resource.Bucket(res["bucket"])
    prefix = res['filePath']
    uploaded_imgs = []
    uploaded_imgs_info = ([], [], [])
    for i in range(start_index, end_index):
        if i >= len_img_paths:
            break
        path = img_paths[i]
        tried_upload[thread_id].append(path)
        try:
            if from_s3_bucket is not None:
                file = io.BytesIO()
                from_s3_object = from_s3.Object(from_s3_bucket, path)
                file_size = from_s3_object.content_length
                if file_size > common.MAX_IMAGE_SIZE:
                    raise SAImageSizeTooLarge(file_size)
                from_s3_object.download_fileobj(file)
            else:
                file_size = Path(path).stat().st_size
                if file_size > common.MAX_IMAGE_SIZE:
                    raise SAImageSizeTooLarge(file_size)
                with open(path, "rb") as f:
                    file = io.BytesIO(f.read())
            images_array = get_image_array_to_upload(
                Path(path).name, file, image_quality_in_editor, project["type"]
            )
            key = upload_image_array_to_s3(bucket, *images_array, prefix)
        except Exception as e:
            logger.warning("Unable to upload image %s. %s", path, e)
            couldnt_upload[thread_id].append(path)
            continue
        else:
            uploaded_imgs.append(path)
            uploaded_imgs_info[0].append(Path(path).name)
            uploaded_imgs_info[1].append(key)
            uploaded_imgs_info[2].append(images_array[2])
            if len(uploaded_imgs) >= 100:
                try:
                    __create_image(
                        uploaded_imgs_info[0],
                        uploaded_imgs_info[1],
                        project,
                        annotation_status,
                        prefix,
                        uploaded_imgs_info[2],
                        project_folder_id,
                        upload_state="Basic"
                    )
                except SABaseException as e:
                    couldnt_upload[thread_id] += uploaded_imgs
                    logger.warning(e)
                else:
                    uploaded[thread_id] += uploaded_imgs
                uploaded_imgs = []
                uploaded_imgs_info = ([], [], [])
    try:
        __create_image(
            uploaded_imgs_info[0],
            uploaded_imgs_info[1],
            project,
            annotation_status,
            prefix,
            uploaded_imgs_info[2],
            project_folder_id,
            upload_state="Basic"
        )
    except SABaseException as e:
        couldnt_upload[thread_id] += uploaded_imgs
        logger.warning(e)
    else:
        uploaded[thread_id] += uploaded_imgs


def __create_image(
    img_names,
    img_paths,
    project,
    annotation_status,
    remote_dir,
    sizes,
    project_folder_id,
    upload_state="Initial"
):
    if len(img_paths) == 0:
        return
    team_id, project_id = project["team_id"], project["id"]
    upload_state_code = common.upload_state_str_to_int(upload_state)
    data = {
        "project_id": str(project_id),
        "team_id": str(team_id),
        "images": [],
        "annotation_status": annotation_status,
        "meta": {},
        "upload_state": upload_state_code
    }
    if project_folder_id is not None:
        data["folder_id"] = project_folder_id
    for img_data, img_path, size in zip(img_names, img_paths, sizes):
        img_name_uuid = Path(img_path).name
        remote_path = remote_dir + f"{img_name_uuid}"
        if upload_state == "External":
            img_name, img_url = img_data
        else:
            img_name, img_url = img_data, remote_path
        data["images"].append({"name": img_name, "path": img_url})
        data["meta"][img_name] = {
            "width": size[0],
            "height": size[1],
            "annotation_json_path": remote_path + ".json",
            "annotation_bluemap_path": remote_path + ".png"
        }

    response = _api.send_request(
        req_type='POST', path='/image/ext-create', json_req=data
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't ext-create image " + response.text
        )


def _get_upload_auth_token(params, project_id):
    response = _api.send_request(
        req_type='GET',
        path=f'/project/{project_id}/sdkImageUploadToken',
        params=params
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't get upload token " + response.text
        )

    res = response.json()
    return res


def _get_boto_session_by_credentials(credentials):
    return boto3.Session(
        aws_access_key_id=credentials['accessKeyId'],
        aws_secret_access_key=credentials['secretAccessKey'],
        aws_session_token=credentials['sessionToken'],
        region_name=credentials['region']
    )


def __tqdm_thread_image_upload(total_num, tried_upload, finish_event):
    _TIME_TO_UPDATE_IN_TQDM = 1
    with tqdm(total=total_num) as pbar:
        while True:
            finished = finish_event.wait(_TIME_TO_UPDATE_IN_TQDM)
            if not finished:
                sum_all = 0
                for i in tried_upload:
                    sum_all += len(i)
                pbar.update(sum_all - pbar.n)
            else:
                pbar.update(total_num - pbar.n)
                break


def get_duplicate_image_names(project_id, team_id, folder_id, image_paths):
    """ Find duplicated images in a folder.
    :param project_id: project_id
    :type project_id: int
    :param team_id: team_id
    :type team_id: int
    :param folder_id: folder_id
    :type folder_id: int
    :param image_paths: list
    :type image_paths: list of str

    :return: list of duplicated images
    :rtype: list of strs
    """
    duplicate_images = []
    image_paths_lists = divide_chunks(image_paths, 500)
    for image_path_list in image_paths_lists:
        json_req = {
            'project_id': project_id,
            'team_id': team_id,
            'folder_id': folder_id
        }
        image_list = [Path(image_path).name for image_path in image_path_list]
        json_req['names'] = image_list
        response = _api.send_request(
            req_type='POST',
            path='/images/getBulk',
            json_req=json_req,
        )
        if not response.ok:
            raise SABaseException(
                response.status_code,
                "Couldn't get image metadata. " + response.text
            )
        duplicate_images += response.json()
    duplicate_images_names = [i['name'] for i in duplicate_images]
    return duplicate_images_names


def _upload_images(
    img_paths, team_id, folder_id, project_id, annotation_status,
    from_s3_bucket, image_quality_in_editor, project, folder_name
):

    _NUM_THREADS = 10
    uploaded = [[] for _ in range(_NUM_THREADS)]
    tried_upload = [[] for _ in range(_NUM_THREADS)]
    couldnt_upload = [[] for _ in range(_NUM_THREADS)]
    finish_event = threading.Event()
    params = {'team_id': team_id, 'folder_id': folder_id}
    res = _get_upload_auth_token(params=params, project_id=project_id)

    prefix = res['filePath']
    limit = res['availableImageCount']
    imgs_to_upload = img_paths[:limit]
    duplicate_images_names = get_duplicate_image_names(
        project_id=project_id,
        team_id=team_id,
        folder_id=folder_id,
        image_paths=imgs_to_upload
    )

    if len(duplicate_images_names):
        logger.warning(
            "%s already existing images found that won't be uploaded.",
            len(duplicate_images_names)
        )
    imgs_to_upload = [
        i for i in imgs_to_upload if Path(i).name not in duplicate_images_names
    ]
    logger.info(
        "Uploading %s images to project %s.", len(imgs_to_upload), folder_name
    )

    images_to_skip = [str(path) for path in img_paths[limit:]]

    chunksize = int(math.ceil(len(imgs_to_upload) / _NUM_THREADS))

    tqdm_thread = threading.Thread(
        target=__tqdm_thread_image_upload,
        args=(len(imgs_to_upload), tried_upload, finish_event),
        daemon=True
    )
    tqdm_thread.start()

    threads = []
    for thread_id in range(_NUM_THREADS):
        t = threading.Thread(
            target=__upload_images_to_aws_thread,
            args=(
                res, imgs_to_upload, project, annotation_status, prefix,
                thread_id, chunksize, couldnt_upload, uploaded, tried_upload,
                image_quality_in_editor, from_s3_bucket, folder_id
            ),
            daemon=True
        )
        threads.append(t)
        t.start()
    for thread in threads:
        thread.join()
    finish_event.set()
    tqdm_thread.join()
    not_uploaded = [str(f) for s in couldnt_upload for f in s]
    uploaded = [str(f) for s in uploaded for f in s]
    not_uploaded += images_to_skip

    return (uploaded, not_uploaded, duplicate_images_names)


def _attach_urls(
    img_names_urls, team_id, folder_id, project_id, annotation_status, project,
    folder_name
):
    _NUM_THREADS = 10
    params = {'team_id': team_id, 'folder_id': folder_id}
    uploaded = [[] for _ in range(_NUM_THREADS)]
    tried_upload = [[] for _ in range(_NUM_THREADS)]
    couldnt_upload = [[] for _ in range(_NUM_THREADS)]
    finish_event = threading.Event()

    res = _get_upload_auth_token(params=params, project_id=project_id)

    prefix = res['filePath']
    limit = res['availableImageCount']
    images_to_upload = img_names_urls[:limit]

    img_names = [i[0] for i in images_to_upload]
    duplicate_images = get_duplicate_image_names(
        project_id=project_id,
        team_id=team_id,
        folder_id=folder_id,
        image_paths=img_names
    )
    if len(duplicate_images) != 0:
        logger.warning(
            "%s already existing images found that won't be uploaded.",
            len(duplicate_images)
        )

    images_to_upload = [
        i for i in images_to_upload if i[0] not in duplicate_images
    ]
    logger.info(
        "Uploading %s images to project %s.", len(images_to_upload), folder_name
    )

    images_to_skip = img_names_urls[limit:]
    chunksize = int(math.ceil(len(images_to_upload) / _NUM_THREADS))

    tqdm_thread = threading.Thread(
        target=__tqdm_thread_image_upload,
        args=(len(images_to_upload), tried_upload, finish_event),
        daemon=True
    )
    tqdm_thread.start()
    threads = []
    for thread_id in range(_NUM_THREADS):
        t = threading.Thread(
            target=__attach_image_urls_to_project_thread,
            args=(
                res, images_to_upload, project, annotation_status, prefix,
                thread_id, chunksize, couldnt_upload, uploaded, tried_upload,
                folder_id
            ),
            daemon=True
        )
        threads.append(t)
        t.start()
    for t in threads:
        t.join()
    finish_event.set()
    tqdm_thread.join()
    list_of_not_uploaded = []
    for couldnt_upload_thread in couldnt_upload:
        for f in couldnt_upload_thread:
            list_of_not_uploaded.append(str(f))
    list_of_uploaded = []
    for upload_thread in uploaded:
        for f in upload_thread:
            list_of_uploaded.append(str(f))

    list_of_not_uploaded += [i[0] for i in images_to_skip]
    return (list_of_uploaded, list_of_not_uploaded, duplicate_images)


def __attach_image_urls_to_project_thread(
    res, img_names_urls, project, annotation_status, prefix, thread_id,
    chunksize, couldnt_upload, uploaded, tried_upload, project_folder_id
):
    len_img_paths = len(img_names_urls)
    start_index = thread_id * chunksize
    end_index = start_index + chunksize
    if start_index >= len_img_paths:
        return
    s3_session = _get_boto_session_by_credentials(res)
    s3_resource = s3_session.resource('s3')
    bucket = s3_resource.Bucket(res["bucket"])
    prefix = res['filePath']
    uploaded_imgs = []
    uploaded_imgs_info = ([], [], [])
    for i in range(start_index, end_index):
        if i >= len_img_paths:
            break
        name, _ = img_names_urls[i]
        tried_upload[thread_id].append(name)
        img_name_hash = str(uuid.uuid4()) + Path(name).suffix
        key = prefix + img_name_hash
        try:
            bucket.put_object(
                Body=json.dumps(create_empty_annotation((None, None), name)),
                Key=key + ".json"
            )
        except Exception as e:
            logger.warning("Unable to upload image %s. %s", name, e)
            couldnt_upload[thread_id].append(name)
            continue
        else:
            uploaded_imgs.append(name)
            uploaded_imgs_info[0].append(img_names_urls[i])
            uploaded_imgs_info[1].append(key)
            uploaded_imgs_info[2].append((None, None))
            if len(uploaded_imgs) >= 100:
                try:
                    __create_image(
                        uploaded_imgs_info[0],
                        uploaded_imgs_info[1],
                        project,
                        annotation_status,
                        prefix,
                        uploaded_imgs_info[2],
                        project_folder_id,
                        upload_state="External"
                    )
                except SABaseException as e:
                    couldnt_upload[thread_id] += uploaded_imgs
                    logger.warning(e)
                else:
                    uploaded[thread_id] += uploaded_imgs
                uploaded_imgs = []
                uploaded_imgs_info = ([], [], [])
    try:
        __create_image(
            uploaded_imgs_info[0],
            uploaded_imgs_info[1],
            project,
            annotation_status,
            prefix,
            uploaded_imgs_info[2],
            project_folder_id,
            upload_state="External"
        )
    except SABaseException as e:
        couldnt_upload[thread_id] += uploaded_imgs
        logger.warning(e)
    else:
        uploaded[thread_id] += uploaded_imgs



from pathlib import Path
import superannotate as sa
from ..common import upload_project


def test_sagemaker_instance_segmentation(tmpdir):
    project_name = "sagemaker_instance_pixel"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'SageMaker' / 'input' / 'toSuperAnnotate' / 'instance_segmentation'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'SageMaker', 'test-obj-detect', 'Pixel',
        'instance_segmentation'
    )

    description = 'sagemaker vector instance segmentation'
    ptype = 'Pixel'
    upload_project(out_dir, project_name, description, ptype)


def test_sagemaker_object_detection(tmpdir):
    project_name = "sagemaker_object_vector"

    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'SageMaker' / 'input' / 'toSuperAnnotate' / 'object_detection'
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, 'SageMaker', 'test-obj-detect', 'Vector',
        'object_detection'
    )

    description = 'sagemaker object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)



import io
import logging
import re
import time
from pathlib import Path

import boto3

from .. import common
from ..api import API
from ..exceptions import SABaseException, SAImageSizeTooLarge
from .images import (
    delete_image, get_image_annotations, get_image_bytes, get_image_metadata,
    get_project_root_folder_id, search_images, set_image_annotation_status,
    upload_image_annotations
)
from .project_api import get_project_and_folder_metadata
from .projects import (
    get_project_default_image_quality_in_editor, _get_available_image_counts
)
from ..mixp.decorators import Trackable
from .utils import _get_upload_auth_token, _get_boto_session_by_credentials, upload_image_array_to_s3, get_image_array_to_upload, __create_image, __copy_images, __move_images, get_project_folder_string

logger = logging.getLogger("superannotate-python-sdk")
_api = API.get_instance()


@Trackable
def upload_image_to_project(
    project,
    img,
    image_name=None,
    annotation_status="NotStarted",
    from_s3_bucket=None,
    image_quality_in_editor=None
):
    """Uploads image (io.BytesIO() or filepath to image) to project.
    Sets status of the uploaded image to set_status if it is not None.

    :param project: project name or folder path (e.g., "project1/folder1")
    :type project: str
    :param img: image to upload
    :type img: io.BytesIO() or Pathlike (str or Path)
    :param image_name: image name to set on platform. If None and img is filepath,
                       image name will be set to filename of the path
    :type image_name: str
    :param annotation_status: value to set the annotation statuses of the uploaded image NotStarted InProgress QualityCheck Returned Completed Skipped
    :type annotation_status: str
    :param from_s3_bucket: AWS S3 bucket to use. If None then folder_path is in local filesystem
    :type from_s3_bucket: str
    :param image_quality_in_editor: image quality be seen in SuperAnnotate web annotation editor.
           Can be either "compressed" or "original".  If None then the default value in project settings will be used.
    :type image_quality_in_editor: str
    """
    initial_project_inp = project
    project, folder = get_project_and_folder_metadata(project)
    upload_state = common.upload_state_int_to_str(project.get("upload_state"))
    if upload_state == "External":
        raise SABaseException(
            0,
            "The function does not support projects containing images attached with URLs"
        )
    annotation_status = common.annotation_status_str_to_int(annotation_status)
    if image_quality_in_editor is None:
        image_quality_in_editor = get_project_default_image_quality_in_editor(
            project
        )

    img_name = None
    if not isinstance(img, io.BytesIO):
        img_name = Path(img).name
        if from_s3_bucket is not None:
            from_session = boto3.Session()
            from_s3 = from_session.resource('s3')
            from_s3_object = from_s3.Object(from_s3_bucket, img)
            file_size = from_s3_object.content_length
            if file_size > common.MAX_IMAGE_SIZE:
                raise SAImageSizeTooLarge(file_size, img)
            img = io.BytesIO()
            from_s3_object.download_fileobj(img)
        else:
            file_size = Path(img).stat().st_size
            if file_size > common.MAX_IMAGE_SIZE:
                raise SAImageSizeTooLarge(file_size, img)
            with open(img, "rb") as f:
                img = io.BytesIO(f.read())
    elif img.getbuffer().nbytes > common.MAX_IMAGE_SIZE:
        raise SAImageSizeTooLarge(img.getbuffer().nbytes)

    if image_name is not None:
        img_name = image_name

    if img_name is None:
        raise SABaseException(
            0, "Image name img_name should be set if img is not Pathlike"
        )

    if folder:
        folder_id = folder["id"]
    else:
        folder_id = get_project_root_folder_id(project)

    team_id, project_id = project["team_id"], project["id"]
    params = {'team_id': team_id, 'folder_id': folder_id}
    res = _get_upload_auth_token(params=params, project_id=project_id)
    prefix = res['filePath']
    s3_session = _get_boto_session_by_credentials(res)
    s3_resource = s3_session.resource('s3')
    bucket = s3_resource.Bucket(res["bucket"])
    try:
        images_info_and_array = get_image_array_to_upload(
            img_name, img, image_quality_in_editor, project["type"]
        )
        key = upload_image_array_to_s3(bucket, *images_info_and_array, prefix)
    except Exception as e:
        raise SABaseException(0, "Couldn't upload to data server.") from e

    __create_image(
        [img_name], [key],
        project,
        annotation_status,
        prefix, [images_info_and_array[2]],
        folder_id,
        upload_state="Basic"
    )

    while True:
        try:
            get_image_metadata(initial_project_inp, img_name)
        except SABaseException:
            time.sleep(0.2)
        else:
            break


def _copy_images(
    source_project, destination_project, image_names, include_annotations,
    copy_annotation_status, copy_pin
):
    NUM_TO_SEND = 500
    source_project, source_project_folder = source_project
    destination_project, destination_project_folder = destination_project
    if source_project["id"] != destination_project["id"]:
        raise SABaseException(
            0,
            "Source and destination projects should be the same for copy_images"
        )
    params = {
        "team_id": source_project["team_id"],
        "project_id": source_project["id"]
    }
    if source_project_folder is not None:
        source_folder_name = source_project_folder["name"]
    else:
        source_folder_name = 'root'
    json_req = {"source_folder_name": source_folder_name}
    if destination_project_folder is not None:
        destination_folder_id = destination_project_folder["id"]
    else:
        destination_folder_id = get_project_root_folder_id(destination_project)
    json_req["destination_folder_id"] = destination_folder_id
    res = {}
    res['skipped'] = []
    res['completed'] = []
    for start_index in range(0, len(image_names), NUM_TO_SEND):
        json_req["image_names"] = image_names[start_index:start_index +
                                              NUM_TO_SEND]
        response = _api.send_request(
            req_type='POST',
            path='/image/copy',
            params=params,
            json_req=json_req
        )

        if not response.ok:
            raise SABaseException(
                response.status_code, "Couldn't copy images " + response.text
            )
        res['skipped'] += response.json()['skipped']
        res['completed'] += response.json()['completed']

    for image_name in image_names:
        _copy_annotations_and_metadata(
            source_project, source_project_folder, image_name,
            destination_project, destination_project_folder, image_name,
            include_annotations, copy_annotation_status, copy_pin
        )
    return res


@Trackable
def copy_images(
    source_project,
    image_names,
    destination_project,
    include_annotations=True,
    copy_annotation_status=True,
    copy_pin=True
):
    """Copy images in bulk between folders in a project 

    :param source_project: project name or folder path (e.g., "project1/folder1")
    :type source_project: str
    :param image_names: image names. If None, all images from source project will be copied
    :type image: list of str
    :param destination_project: project name or folder path (e.g., "project1/folder2")
    :type destination_project: str
    :param include_annotations: enables annotations copy
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    :return: list of skipped image names
    :rtype: list of strs
    """

    source_project_inp = source_project
    destination_project_inp = destination_project

    source_project, source_folder = get_project_and_folder_metadata(
        source_project
    )
    destination_project, destination_folder = get_project_and_folder_metadata(
        destination_project
    )
    root_folder_id = get_project_root_folder_id(source_project)

    destination_folder_id = root_folder_id
    source_folder_id = root_folder_id

    if destination_folder:
        destination_folder_id = destination_folder['id']
    if source_folder:
        source_folder_id = source_folder['id']

    if image_names == None:
        image_names = search_images(source_project_inp)

    done_count, total_skipped_list, logs = __copy_images(
        source_project, source_folder_id, destination_folder_id, image_names,
        include_annotations, copy_pin
    )
    for log in logs:
        logger.info(log)

    if done_count > 1 or done_count == 0:
        message = f"Copied {done_count}/{len(image_names)} images from {get_project_folder_string(source_project_inp)} to {get_project_folder_string(destination_project_inp)}."
        logger.info(message)

    elif done_count == 1:
        message = f"Copied an image from {get_project_folder_string(source_project_inp)} to {get_project_folder_string(destination_project_inp)}."
        logger.info(message)

    return total_skipped_list


@Trackable
def delete_images(project, image_names):
    """Delete images in project.

    :param project: project name or folder path (e.g., "project1/folder1")
    :type project: str
    :param image_names: to be deleted images' names. If None, all the images will be deleted
    :type image_names: list of strs
    """
    NUM_TO_SEND = 1000
    project, project_folder = get_project_and_folder_metadata(project)
    params = {"team_id": project["team_id"], "project_id": project["id"]}
    if image_names is None:
        images = search_images((project, project_folder), return_metadata=True)
    else:
        if not isinstance(image_names, list):
            raise SABaseException(
                0, "image_names should be a list of strs or None"
            )
        images = get_image_metadata(
            (project, project_folder),
            image_names,
            return_dict_on_single_output=False
        )
    for start_index in range(0, len(images), NUM_TO_SEND):
        data = {
            "image_ids":
                [
                    image["id"]
                    for image in images[start_index:start_index + NUM_TO_SEND]
                ]
        }
        response = _api.send_request(
            req_type='PUT',
            path='/image/delete/images',
            params=params,
            json_req=data
        )
        if not response.ok:
            raise SABaseException(
                response.status_code, "Couldn't delete images " + response.text
            )
    logger.info(
        "Images deleted in project %s%s", project["name"],
        "" if project_folder is None else "/" + project_folder["name"]
    )


@Trackable
def move_images(
    source_project,
    image_names,
    destination_project,
    include_annotations=True,
    copy_annotation_status=True,
    copy_pin=True,
):
    """Move images in bulk between folders in a project 

    :param source_project: project name or folder path (e.g., "project1/folder1")
    :type source_project: str
    :param image_names: image names. If None, all images from source project will be moved
    :type image: list of str
    :param destination_project: project name or folder path (e.g., "project1/folder2")
    :type destination_project: str
    :param include_annotations: enables annotations copy
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    :return: list of skipped image names
    :rtype: list of strs
    """
    source_project_inp = source_project
    destination_project_inp = destination_project

    source_project, source_folder = get_project_and_folder_metadata(
        source_project
    )
    destination_project, destination_folder = get_project_and_folder_metadata(
        destination_project
    )
    root_folder_id = get_project_root_folder_id(source_project)

    destination_folder_id = root_folder_id
    source_folder_id = root_folder_id

    if destination_folder:
        destination_folder_id = destination_folder['id']
    if source_folder:
        source_folder_id = source_folder['id']

    if image_names == None:
        image_names = search_images(source_project_inp)

    moved, skipped, logs = __move_images(
        source_project, source_folder_id, destination_folder_id, image_names
    )

    for log in logs:
        logger.info(log)

    if len(moved) > 1 or len(moved) == 0:
        message = f"Moved {len(moved)}/{len(image_names)} images from {get_project_folder_string(source_project_inp)} to {get_project_folder_string(destination_project_inp)}."
        logger.info(message)
    elif len(moved) == 1:
        message = f"Moved an image from {get_project_folder_string(source_project_inp)} to {get_project_folder_string(destination_project_inp)}."
        logger.info(message)

    return skipped


@Trackable
def copy_image(
    source_project,
    image_name,
    destination_project,
    include_annotations=False,
    copy_annotation_status=False,
    copy_pin=False
):
    """Copy image to a project. The image's project is the same as destination
    project then the name will be changed to <image_name>_(<num>).<image_ext>,
    where <num> is the next available number deducted from project image list.

    :param source_project: project name plus optional subfolder in the project (e.g., "project1/folder1") or
                           metadata of the project of source project
    :type source_project: str or dict
    :param image_name: image name
    :type image: str
    :param destination_project: project name or metadata of the project of destination project
    :type destination_project: str or dict
    :param include_annotations: enables annotations copy
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    """
    source_project, source_project_folder = get_project_and_folder_metadata(
        source_project
    )
    destination_project, destination_project_folder = get_project_and_folder_metadata(
        destination_project
    )
    img_b = get_image_bytes((source_project, source_project_folder), image_name)
    new_name = image_name
    extension = Path(image_name).suffix
    p = re.compile(r"_\([0-9]+\)\.")
    while True:
        try:
            get_image_metadata(
                (destination_project, destination_project_folder), new_name
            )
        except SABaseException:
            break
        else:
            for m in p.finditer(new_name):
                if m.start() + len(m.group()
                                  ) + len(extension) - 1 == len(new_name):
                    num = int(m.group()[2:-2])
                    new_name = new_name[:m.start() +
                                        2] + str(num + 1) + ")" + extension
                    break
            else:
                new_name = Path(new_name).stem + "_(1)" + extension

    upload_image_to_project(
        (destination_project, destination_project_folder), img_b, new_name
    )
    _copy_annotations_and_metadata(
        source_project, source_project_folder, image_name, destination_project,
        destination_project_folder, new_name, include_annotations,
        copy_annotation_status, copy_pin
    )
    logger.info(
        "Copied image %s/%s to %s/%s.", source_project["name"], image_name,
        destination_project["name"], new_name
    )


def _copy_annotations_and_metadata(
    source_project, source_project_folder, image_name, destination_project,
    destination_project_folder, new_name, include_annotations,
    copy_annotation_status, copy_pin
):
    if include_annotations:
        annotations = get_image_annotations(
            (source_project, source_project_folder), image_name
        )
        if annotations["annotation_json"] is not None:
            if "annotation_mask" in annotations:
                if annotations["annotation_mask"] is not None:
                    upload_image_annotations(
                        (destination_project, destination_project_folder),
                        new_name, annotations["annotation_json"],
                        annotations["annotation_mask"]
                    )
            else:
                upload_image_annotations(
                    (destination_project, destination_project_folder), new_name,
                    annotations["annotation_json"]
                )
    if copy_annotation_status or copy_pin:
        img_metadata = get_image_metadata(
            (source_project, source_project_folder), image_name
        )
        if copy_annotation_status:
            set_image_annotation_status(
                (destination_project, destination_project_folder), new_name,
                img_metadata["annotation_status"]
            )
        if copy_pin:
            pin_image(
                (destination_project, destination_project_folder), new_name,
                img_metadata["is_pinned"]
            )


@Trackable
def move_image(
    source_project,
    image_name,
    destination_project,
    include_annotations=True,
    copy_annotation_status=True,
    copy_pin=True
):
    """Move image from source_project to destination_project. source_project
    and destination_project cannot be the same.

    :param source_project: project name or metadata of the project of source project
    :type source_project: str or dict
    :param image_name: image name
    :type image: str
    :param destination_project: project name or metadata of the project of destination project
    :type destination_project: str or dict
    :param include_annotations: enables annotations move
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    """
    source_project, source_project_folder = get_project_and_folder_metadata(
        source_project
    )
    destination_project, destination_project_folder = get_project_and_folder_metadata(
        destination_project
    )
    if source_project == destination_project:
        raise SABaseException(
            0, "Cannot move image if source_project == destination_project."
        )
    copy_image(
        (source_project, source_project_folder), image_name,
        (destination_project, destination_project_folder), include_annotations,
        copy_annotation_status, copy_pin
    )
    delete_image((source_project, source_project_folder), image_name)
    logger.info("Deleted image %s/%s.", source_project["name"], image_name)


@Trackable
def pin_image(project, image_name, pin=True):
    """Pins (or unpins) image

    :param project: project name or folder path (e.g., "project1/folder1")
    :type project: str
    :param image_name: image name
    :type image: str
    :param pin: sets to pin if True, else unpins image
    :type pin: bool
    """
    project, project_folder = get_project_and_folder_metadata(project)
    img_metadata = get_image_metadata((project, project_folder), image_name)
    team_id, project_id, image_id = project["team_id"], project[
        "id"], img_metadata["id"]
    params = {"team_id": team_id, "project_id": project_id}
    if project_folder is not None:
        params["folder_id"] = project_folder["id"]
    json_req = {"is_pinned": int(pin)}
    response = _api.send_request(
        req_type='PUT',
        path=f'/image/{image_id}',
        params=params,
        json_req=json_req
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't pin image " + response.text
        )


@Trackable
def assign_images(project, image_names, user):
    """Assigns images to a user. The assignment role, QA or Annotator, will
    be deduced from the user's role in the project. With SDK, the user can be
    assigned to a role in the project with the share_project function.

    :param project: project name or folder path (e.g., "project1/folder1")
    :type project: str
    :param image_names: list of image names to assign
    :type image_names: list of str
    :param user: user email
    :type user: str
    """
    logger.info("Assign %s images to user %s", len(image_names), user)
    if len(image_names) == 0:
        return
    project, project_folder = get_project_and_folder_metadata(project)
    images = search_images((project, project_folder), return_metadata=True)
    image_dict = {}
    for image in images:
        image_dict[image["name"]] = image["id"]

    image_ids = []
    for image_name in image_names:
        image_ids.append(image_dict[image_name])
    team_id, project_id = project["team_id"], project["id"]
    params = {"team_id": team_id, "project_id": project_id}
    if project_folder is not None:
        params['folder_id'] = project_folder['id']
    json_req = {"user_id": user, "image_ids": image_ids}
    response = _api.send_request(
        req_type='POST',
        path='/images/assign',
        params=params,
        json_req=json_req
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't assign images " + response.text
        )



'''
COCO to SA conversion methods
'''
import logging
import threading
import json
from pathlib import Path

import cv2
import numpy as np

from ....common import write_to_json, tqdm_converter
from ..sa_json_helper import _create_sa_json, _create_vector_instance
from .coco_api import _maskfrRLE, decode

logger = logging.getLogger("superannotate-python-sdk")


def annot_to_polygon(annot):
    if isinstance(annot['counts'], list):
        bitmask = _maskfrRLE(annot)
    elif isinstance(annot['counts'], str):
        bitmask = decode(annot)

    contours, _ = cv2.findContours(
        bitmask.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    segments = []
    for contour in contours:
        contour = contour.flatten().tolist()
        if len(contour) > 4:
            segments.append(contour)

    return segments


def save_sa_jsons(coco_json, img_id_to_annot, output_dir):
    images_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(coco_json["images"]), images_converted, [], finish_event),
        daemon=True
    )
    logger.info('Writting to disk')
    tqdm_thread.start()
    for img in coco_json['images']:
        if 'file_name' in img:
            image_path = Path(img['file_name']).name
        else:
            image_path = img['coco_url'].split('/')[-1]

        if str(img['id']) not in img_id_to_annot:
            sa_instances = []
        else:
            sa_instances = img_id_to_annot[str(img['id'])]
        file_name = "%s___objects.json" % image_path

        sa_metadata = {
            'name': image_path,
            'width': img['width'],
            'height': img['height']
        }
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
    finish_event.set()
    tqdm_thread.join()


def coco_instance_segmentation_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    instance_groups = {}
    for annot in coco_json['annotations']:
        if 'id' in annot:
            if annot['id'] in instance_groups:
                instance_groups[annot['id']] += 1
            else:
                instance_groups[annot['id']] = 1

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(coco_json["annotations"]), annotations_processed, [],
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for annot in coco_json['annotations']:
        if isinstance(annot['segmentation'], dict):
            annot['segmentation'] = annot_to_polygon(annot['segmentation'])

        cat = cat_id_to_cat[annot['category_id']]
        groupid = 0
        for polygon in annot['segmentation']:
            if 'id' in annot and annot[
                'id'] in instance_groups and instance_groups[annot['id']] > 1:
                groupid = annot['id']

            sa_obj = _create_vector_instance(
                'polygon', polygon, {}, [], cat['name']
            )
            if groupid != 0:
                sa_obj['groupId'] = groupid

            if str(annot['image_id']) not in image_id_to_annotations:
                image_id_to_annotations[str(annot['image_id'])] = [sa_obj]
            else:
                image_id_to_annotations[str(annot['image_id'])].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)


def coco_object_detection_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(coco_json["annotations"]), annotations_processed, [],
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for annot in coco_json['annotations']:
        if isinstance(annot['segmentation'], dict):
            annot['segmentation'] = annot_to_polygon(annot['segmentation'])
        cat = cat_id_to_cat[annot['category_id']]

        points = (
            annot['bbox'][0], annot['bbox'][1],
            annot['bbox'][0] + annot['bbox'][2],
            annot['bbox'][1] + annot['bbox'][3]
        )

        sa_obj = _create_vector_instance('bbox', points, {}, [], cat['name'])

        if str(annot['image_id']) not in image_id_to_annotations:
            image_id_to_annotations[str(annot['image_id'])] = [sa_obj]
        else:
            image_id_to_annotations[str(annot['image_id'])].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)


def coco_keypoint_detection_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))

    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = {
            "name": cat["name"],
            "keypoints": cat["keypoints"],
            "skeleton": cat["skeleton"]
        }

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(coco_json["annotations"]), annotations_processed, [],
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for annot in coco_json['annotations']:
        if annot['num_keypoints'] > 0:
            sa_points = [
                item for index, item in enumerate(annot['keypoints'])
                if (index + 1) % 3 != 0
            ]

            sa_points = [
                (sa_points[i], sa_points[i + 1])
                for i in range(0, len(sa_points), 2)
            ]
            if annot["category_id"] in cat_id_to_cat.keys():
                keypoint_names = cat_id_to_cat[annot["category_id"]
                                              ]['keypoints']

                bad_points = []
                id_mapping = {}
                index = 1
                points = []
                for point_index, point in enumerate(sa_points):
                    if sa_points[point_index] == (0, 0):
                        bad_points.append(point_index + 1)
                        continue
                    id_mapping[point_index + 1] = index
                    points.append({'id': index, 'x': point[0], 'y': point[1]})
                    index += 1

                connections = []
                for connection in cat_id_to_cat[annot["category_id"]
                                               ]['skeleton']:

                    index = cat['skeleton'].index(connection)
                    from_point = cat_id_to_cat[annot["category_id"]
                                              ]['skeleton'][index][0]
                    to_point = cat_id_to_cat[annot["category_id"]
                                            ]['skeleton'][index][1]

                    if from_point in bad_points or to_point in bad_points:
                        continue

                    connections.append(
                        {
                            'id': index + 1,
                            'from': id_mapping[from_point],
                            'to': id_mapping[to_point]
                        }
                    )

                pointLabels = {}
                for kp_index, kp_name in enumerate(keypoint_names):
                    if kp_index + 1 in bad_points:
                        continue
                    pointLabels[id_mapping[kp_index + 1] - 1] = kp_name

                sa_obj = _create_vector_instance(
                    'template', points, pointLabels, [],
                    cat_id_to_cat[annot["category_id"]]["name"], connections
                )
                if str(annot['image_id']) not in image_id_to_annotations:
                    image_id_to_annotations[str(annot['image_id'])] = [sa_obj]
                else:
                    image_id_to_annotations[str(annot['image_id']
                                               )].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)



import sys
from .app import mp, get_default
from superannotate.api import API
from .utils import parsers
from threading import Lock
from functools import wraps

_api = API.get_instance()
always_trackable_func_names = ["upload_images_from_folder_to_project"]


class Trackable(object):
    registered = set('<module>')

    def __init__(self, function):
        lock = Lock()
        self.function = function
        with lock:
            Trackable.registered.add(function.__name__)

    def __call__(self, *args, **kwargs):
        try:
            func_name_to_track = self.function.__name__
            caller_name = sys._getframe(1).f_code.co_name
            if caller_name not in Trackable.registered or func_name_to_track in always_trackable_func_names:
                data = getattr(parsers, func_name_to_track)(*args, **kwargs)
                user_id = _api.user_id
                event_name = data['event_name']
                properties = data['properties']
                default = get_default(
                    _api.team_name,
                    _api.user_id,
                    project_name=properties.get('project_name', None)
                )
                properties.pop("project_name", None)
                properties = {**default, **properties}
                if "pytest" not in sys.modules:
                    mp.track(user_id, event_name, properties)
        except:
            pass
        return self.function(*args, **kwargs)



import logging

from ..api import API
from ..exceptions import SABaseException

logger = logging.getLogger("superannotate-python-sdk")
from ..mixp.decorators import Trackable

_api = API.get_instance()


@Trackable
def search_team_contributors(
    email=None, first_name=None, last_name=None, return_metadata=False
):
    """Search for contributors in the team

    :param email: filter by email
    :type email: str
    :param first_name: filter by first name
    :type first_name: str
    :param last_name: filter by last name
    :type last_name: str

    :return: metadata of found users
    :rtype: list of dicts
    """
    result_list = []
    params = {'team_id': _api.team_id, 'offset': 0}
    if email is not None:
        params['email'] = email
    if first_name is not None:
        params['first_name'] = first_name
    if last_name is not None:
        params['last_name'] = last_name

    while True:
        response = _api.send_request(
            req_type='GET', path='/users', params=params
        )
        if response.ok:
            res = response.json()
            result_list += res["data"]
            new_len = len(result_list)
            if res["count"] <= new_len:
                break
            params["offset"] = new_len
        else:
            raise SABaseException(
                response.status_code,
                "Couldn't search projects. " + response.text
            )

    if return_metadata:
        return result_list
    else:
        return [x["email"] for x in result_list]


def get_team_contributor_metadata(email):
    users = search_team_contributors(email, return_metadata=True)
    results = []
    for user in users:
        if user["email"] == email:
            results.append(user)

    if len(results) == 0:
        raise SABaseException(0, "No user with email " + email + " found.")
    if len(results) > 1:
        raise SABaseException(0, "Email " + email + " malformed.")
    return results[0]



import superannotate as sa
from pathlib import Path
import json


def test_json_version_conversion(tmpdir):
    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'sa_json_versions' / 'version2'
    temp_dir = Path(tmpdir) / 'tmp_dir'
    output_dir = Path(tmpdir) / 'output_dir'

    converted_files_old = sa.convert_json_version(input_dir, temp_dir, 1)
    converted_files_new = sa.convert_json_version(temp_dir, output_dir, 2)

    assert len(converted_files_old) == len(converted_files_new)
    files_list = input_dir.glob('*.json')

    metadata_keys = ['height', 'width', 'status', 'pinned']
    comments_keys = ['x', 'y', 'resolved']
    for file in files_list:
        input_data = json.load(open(file))
        output_data = json.load(open(output_dir / file.name))
        for key in metadata_keys:
            assert input_data['metadata'][key] == output_data['metadata'][key]

        assert len(input_data['instances']) == len(output_data['instances'])

        assert len(input_data['comments']) == len(output_data['comments'])
        for in_com, out_com in zip(
            input_data['comments'], output_data['comments']
        ):
            assert len(in_com['correspondence']) == len(
                out_com['correspondence']
            )
            for key in comments_keys:
                assert in_com[key] == out_com[key]

        assert len(input_data['tags']) == len(output_data['tags'])
        for in_tag, out_tag in zip(input_data['tags'], output_data['tags']):
            assert in_tag == out_tag



'''
Labelbox to SA conversion method
'''
import logging
import threading
from pathlib import Path
import cv2
import numpy as np

from .labelbox_helper import (
    image_downloader, _create_classes_id_map, _create_attributes_list
)

from ..sa_json_helper import (_create_pixel_instance, _create_sa_json)

from ....common import (
    hex_to_rgb, blue_color_generator, write_to_json, tqdm_converter
)

logger = logging.getLogger("superannotate-python-sdk")


def labelbox_instance_segmentation_to_sa_pixel(
    json_data, output_dir, input_dir
):
    classes = _create_classes_id_map(json_data)

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_data), images_converted, images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for data in json_data:
        file_name = data['External ID'] + '___pixel.json'
        mask_name = data['External ID'] + '___save.png'
        sa_metadata = {'name': data['External ID']}
        if 'objects' not in data['Label'].keys():
            sa_json = _create_sa_json([], sa_metadata)
            write_to_json(output_dir / file_name, sa_json)
            continue

        instances = data["Label"]["objects"]
        sa_instances = []
        blue_colors = blue_color_generator(len(instances))

        for i, instance in enumerate(instances):
            class_name = instance["value"]
            attributes = []
            if 'classifications' in instance.keys():
                attributes = _create_attributes_list(
                    instance['classifications']
                )

            if 'bbox' in instance.keys() or 'polygon' in instance.keys(
            ) or 'line' in instance.keys() or 'point' in instance.keys():
                continue

            bitmask_name = '%s.png' % instance['featureId']
            downloaded = image_downloader(instance['instanceURI'], bitmask_name)
            if downloaded:
                mask = cv2.imread(bitmask_name)
            else:
                mask = cv2.imread(str(input_dir / 'bitmasks' / bitmask_name))
                bitmask_name = output_dir / bitmask_name

            if isinstance(mask, type(None)):
                logger.warning("Can't open '%s' bitmask.", bitmask_name)
                images_not_converted.append(bitmask_name)
                continue

            if i == 0:
                H, W, _ = mask.shape
                sa_metadata['width'] = W
                sa_metadata['height'] = H
                sa_mask = np.zeros((H, W, 4))
            sa_mask[np.all(mask == [255, 255, 255], axis=2)
                   ] = list(hex_to_rgb(blue_colors[i]))[::-1] + [255]

            parts = [{'color': blue_colors[i]}]
            sa_obj = _create_pixel_instance(parts, attributes, class_name)

            sa_instances.append(sa_obj.copy())
            Path(bitmask_name).unlink()

        images_converted.append(data['External ID'])
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
        cv2.imwrite(str(output_dir / mask_name), sa_mask)

    finish_event.set()
    tqdm_thread.join()
    return classes



import json

from ..baseStrategy import baseStrategy

from ....common import write_to_json


class SuperviselyStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        sa_classes, classes_id_map = self._create_sa_classes()
        json_files = []
        if self.dataset_name != '':
            json_files.append(
                self.export_root / 'ds' / 'ann' / (self.dataset_name + '.json')
            )
        else:
            files_gen = (self.export_root / 'ds' / 'ann').glob('*')
            json_files = list(files_gen)

        if self.conversion_algorithm.__name__ == 'supervisely_keypoint_detection_to_sa_vector':
            meta_json = json.load(open(self.export_root / 'meta.json'))
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, meta_json, self.output_dir
            )
        elif self.conversion_algorithm.__name__ == 'supervisely_instance_segmentation_to_sa_pixel':
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, self.output_dir
            )
        else:
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, self.task, self.output_dir
            )
        (self.output_dir / 'classes').mkdir(exist_ok=True)
        write_to_json(self.output_dir / 'classes' / 'classes.json', sa_classes)

    def _create_sa_classes(self):
        classes_json = json.load(open(self.export_root / 'meta.json'))

        attributes = []
        for tag in classes_json['tags']:
            attributes.append({'name': tag['name']})

        classes_id_map = {}
        classes_loader = []
        for class_ in classes_json['classes']:
            group_name = 'converted_attributs'
            classes_id_map[class_['title']] = {
                'attr_group': {
                    'group_name': group_name,
                    'attributes': []
                }
            }
            for attribute in attributes:
                attribute['groupName'] = group_name
                classes_id_map[class_['title']
                              ]['attr_group']['attributes'].append(
                                  {'name': attribute['name']}
                              )

            if not attributes:
                attribute_groups = []
            else:
                attribute_groups = [
                    {
                        'name': group_name,
                        'is_multiselect': 1,
                        'attributes': attributes
                    }
                ]

            attr_group = {
                'name': class_['title'],
                'color': class_['color'],
                'attribute_groups': attribute_groups
            }
            classes_loader.append(attr_group)
        return classes_loader, classes_id_map



'''
VGG to SA conversion method.
'''
import json
import logging
import threading
import cv2
from .vgg_helper import _create_attribute_list
from ..sa_json_helper import (_create_vector_instance, _create_sa_json)

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def vgg_to_sa(json_data, task, output_dir):
    images = json.load(open(json_data))
    if task == 'object_detection':
        instance_types = ['rect']
    elif task == 'instance_segmentation':
        instance_types = ['polygon']
    elif task == 'vector_annotation':
        instance_types = [
            'rect', 'polygon', 'polyline', 'point', 'ellipse', 'circle'
        ]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(images.items()), images_converted, images_not_converted,
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()

    class_id_map = {}
    for _, img in images.items():
        try:
            H, W, _ = cv2.imread(str(output_dir / img['filename'])).shape
        except Exception as e:
            logger.warning(
                "Can't open %s image. 'height' and 'width' for SA JSON metadata will set to zero",
                img['filename']
            )
            H = 0
            W = 0

        file_name = '%s___objects.json' % img['filename']
        sa_metadata = {'name': img['filename'], 'width': W, 'height': H}
        sa_instances = []
        instances = img['regions']
        for instance in instances:
            if 'type' not in instance['region_attributes'].keys():
                raise KeyError(
                    "'VGG' JSON should contain 'type' key which will \
                    be category name. Please correct JSON file."
                )
            if not isinstance(instance['region_attributes']['type'], str):
                raise ValueError(
                    "Wrong attribute was choosen for 'type' attribute."
                )

            class_name = instance['region_attributes']['type']
            if class_name not in class_id_map.keys():
                class_id_map[class_name] = {'attribute_groups': {}}

            attributes = _create_attribute_list(
                instance['region_attributes'], class_name, class_id_map
            )

            if instance['shape_attributes']['name'] in instance_types:
                if instance['shape_attributes']['name'] == 'polygon' or instance[
                    'shape_attributes']['name'] == 'polyline':
                    points = []
                    for x, y in zip(
                        instance['shape_attributes']['all_points_x'],
                        instance['shape_attributes']['all_points_y']
                    ):
                        points.append(x)
                        points.append(y)
                    instance_type = instance['shape_attributes']['name']
                elif instance['shape_attributes']['name'] == 'rect':
                    points = (
                        instance['shape_attributes']['x'],
                        instance['shape_attributes']['y'],
                        instance['shape_attributes']['x'] +
                        instance['shape_attributes']['width'],
                        instance['shape_attributes']['y'] +
                        instance['shape_attributes']['height']
                    )
                    instance_type = 'bbox'
                elif instance['shape_attributes']['name'] == 'ellipse':
                    points = (
                        instance['shape_attributes']['cx'],
                        instance['shape_attributes']['cy'],
                        instance['shape_attributes']['rx'],
                        instance['shape_attributes']['ry'],
                        instance['shape_attributes']['theta']
                    )
                    instance_type = 'ellipse'
                elif instance['shape_attributes']['name'] == 'circle':
                    points = (
                        instance['shape_attributes']['cx'],
                        instance['shape_attributes']['cy'],
                        instance['shape_attributes']['r'],
                        instance['shape_attributes']['r'], 0
                    )
                    instance_type = 'ellipse'
                elif instance['shape_attributes']['name'] == 'point':
                    points = (
                        instance['shape_attributes']['cx'],
                        instance['shape_attributes']['cy']
                    )
                    instance_type = 'point'
                sa_obj = _create_vector_instance(
                    instance_type, points, {}, attributes, class_name
                )
                sa_instances.append(sa_obj)
        images_converted.append(img['filename'])
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()
    return class_id_map
